"""A simple command-line user interface for controlling the pump."""

import ast
import inspect
from pathlib import Path
# Importing readline adds command editing etc. to the input function.
import readline  # pylint: disable=unused-import
import sys
import textwrap

from turboctl import __version__
from turboctl.telegram.parser import PARAMETERS, ERRORS, WARNINGS
from turboctl.ui.control_interface import ControlInterface
from turboctl.ui.docs import docs
from turboctl.ui.table import table


class CommandLineUI:
    """A simple command-line UI.

    Instances of this class should be closed after they are no longer
    needed by calling :attr:`control_interface`\
:meth:`.close() <turboctl.ui.control_interface.ControlInterface.close>`
    or using a ``with`` block. Otherwise the parallel thread created by
    :attr:`control_interface` may continue to run in the background and
    consume resources.

    This class contains several methods with names following the pattern
    ``cmd_<command>``.
    These methods correspond to the commands a user can give to the
    UI: when the command string ``'example'`` is entered, the ``cmd_example()``
    method is called.
    Because messages printed by the ``help`` command use the
    docstrings of these methods in an unaltered form, the docstrings
    are written in plain text and cannot contain any reStructuredText
    syntax that isn't easily readable by humans.

    Attributes:
        inputfile:
            A file-like object for receiving input.

        outputfile:
            A file-like object for writing output.

        debug (bool):
            Setting this to ``True`` activates the debug mode;
            the default value is ``False``.
            See :meth:`cmd_debug` for details.
            
        verbose (bool):
            Setting this to ``True`` increases the verbosity of the output of
            commands; the default value is ``False``.
            See :meth:`cmd_verbose` for details.

        control_interface:
            An :class:`~turboctl.ui.control_interface.ControlInterface` object
            used for sending commands to the pump.

        intro:
            A string displayed to the user upon starting the UI.
            The default value is ``f"Welcome to TurboCtl v. {__version__}! Type 'help' for a list of commands."``.

        prompt:
            A string displayed to the user every time the UI
            requests input. The default value is ``'>> '``.

        indent:
            Class attribute.
            A string used to indent text blocks.
            The default value is ``'    '`` (four spaces).

        cmds_and_aliases:
            Class attribute.
            A list of tuples, where index 0 of each tuple contains
            the name of a command accepted by the UI, and index 1
            contains a list of aliases for that command.
            It has the following value:
            ::

                [
                    ('pump'   , []),
                    ('status' , ['s']),
                    ('reset'  , ['re']),
                    ('read'   , ['r']),
                    ('write'  , ['w']),
                    ('info'   , ['i']),
                    ('exit'   , ['e', 'q', 'x']),
                    ('help'   , ['h']),
                    ('docs'   , ['d']),        
                    ('debug'  , ['db']),
                    ('verbose', ['v']),
                ]
    """

    indent = '    '

    # pylint: disable=bad-whitespace
    cmds_and_aliases = [
        ('pump'   , []),
        ('status' , ['s']),
        ('reset' ,  ['re']),
        ('read'   , ['r']),
        ('write'  , ['w']),
        ('info'   , ['i']),
        ('exit'   , ['e', 'q', 'x']),
        ('help'   , ['h']),
        ('docs'   , ['d']),        
        ('debug'  , ['db']),
        ('verbose', ['v'])
    ]
    # pylint: enable=bad-whitespace
    
    _dicts = {'p': PARAMETERS, 'e': ERRORS, 'w': WARNINGS}
    _widths = {'p': {'name': 20, 'type': 16, 'description': 50}, 
               'e': {'name': 30, 'possible_cause': 50, 'remedy': 50}, 
               'w': {'name': 30, 'possible_cause': 50, 'remedy': 50}
    }
    """A dict of *width* arguments for table.table."""

    def __init__(self, port=None, auto_update=True, 
                 inputfile=None, outputfile=None):
        """Initialize a new CommandLineUI.

        Args:
            port:
                This is passed to the initializer of :attr:`control_interface`.
            auto_update:
                This is also passed to :attr:`control_interface`.
            inputfile:
                The value of :attr:`inputfile`.
                If this is ``None``, the value is :data:`sys.stdin`.
            outputfile:
                The value of :attr:`outputfile`.
                If this is ``None``, the value is :data:`sys.stdout`.
        """
        if inputfile:
            self.inputfile = inputfile
        else:
            self.inputfile = sys.stdin

        if outputfile:
            self.outputfile = outputfile
        else:
            self.outputfile = sys.stdout

        # Replacing sys.stdin and sys.stdout would be simpler, but
        # could lead to unintended consequences, and would e.g.
        # prevent using multiple CommandLineUIs at once.

        self.control_interface = ControlInterface(port, auto_update)
        self.debug = False
        self.verbose = False
        self.intro = (f"Welcome to TurboCtl v. {__version__}! "
                      f"Type 'help' for a list of commands.")
        self.prompt = '>> '
        self._stop_flag = False

    def __enter__(self):
        """Called upon entering a ``with`` block; returns *self*."""
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """Called upon exiting a ``with`` block;
        calls :attr:`control_interface`:meth:`.close()
        <turboctl.ui.control_interface.ControlInterface.close()>`.

        The arguments are ignored.
        """
        self.control_interface.close()

    def run(self):
        """Run the UI.

        This function runs the following algorithm:

        1. Print :attr:`prompt` and wait for user input.
        2. Parse the input string into a command and its arguments by
           splitting it at whitespace.
        3. Call a method with the name ``cmd_<the given command>``.
           The method will execute the command and print its output.
           If no method with that name exists, print an error message
           and start again from step 1.
        4. If the method was :meth:`cmd_exit`, break the loop.
           Otherwise, start another iteration from step 1.
        """
        self.print(self.intro)

        while not self._stop_flag:
            string = self.input(self.prompt)
            self._run_command(string)

    # Don't use reStructuredText in command docstrings, since they are
    # displayed to the user unaltered.

    def cmd_pump(self, value):
        """Turn the pump on or off.

        Values of '1', 'True' and 'on' turn the pump on;
        '0', 'False' and 'off' turn it off.
        """
        # 1 == True and 0 == False.
        if value in [True, 'on']:
            self.print('Turning the pump on')
            return self.control_interface.pump_on()

        if value in [False, 'off']:
            self.print('Turning the pump off')
            return self.control_interface.pump_off()

        raise ValueError('invalid value')

    def cmd_status(self):
        """Get the status of the pump."""
        query, reply = self.control_interface.get_status()

        descriptions = (member.description for member in reply.flag_bits)        
        lines = (self.indent + s for s in descriptions)
        
        if lines:
            condition_str = 'Active status conditions:\n' + '\n'.join(lines)
        else:
            condition_str = 'No active status conditions'

        hardware_str = (
            f'Frequency: {reply.frequency} Hz\n'
            f'Temperature: {reply.temperature} °C\n'
            f'Current: {0.1 * reply.current:.1f} A\n'
            f'Voltage: {reply.voltage} V'
        )
        
        string = textwrap.indent(condition_str + '\n' + hardware_str,
                                 self.indent)
        
        self.print('Pump status:\n' + string)
        return query, reply

    def cmd_reset(self):
        """Reset the error status of the pump."""
        self.print('Resetting error status')
        return self.control_interface.get_status()

    def cmd_read(self, number, index=0):
        """Return the value of parameter *number*, index *index*."""        
        query, reply = self.control_interface.read_parameter(number, index)

        self._print_parameter_output(reply)
        return query, reply

    def cmd_write(self, number, value, index=0):
        """Write *value* to parameter *number*, index *index*."""        
        query, reply = self.control_interface.write_parameter(
            number, value, index)
        
        self._print_parameter_output(reply)
        return query, reply
    
    def _print_parameter_output(self, reply):
        """Format and print out the parameter information in *reply*."""
        
        error = reply.parameter_error
        if error is not None:
            self.print(f'Error: {error.description}')
            return
        
        # Truncate long floats.
        if isinstance(reply.parameter_value, float):
            value_str = f'{reply.parameter_value:.2f}'
        else:
            value_str = str(reply.parameter_value)
            
        self.print(
            f'The value of parameter {reply.parameter_number}, '
            f'index {reply.parameter_index} is {value_str}'
        )

    def cmd_info(self, letter, numbers):
        """Display information about parameters, errors, or warnings.
        
        *letter* should be 'p', 'e', or 'w' depending on whether *numbers*
        refer to parameters, errors, or warnings.
        
        *numbers* should be a list or a tuple of the numbers of those
        parameters/errors/warnings that should be displayed. It can also be a
        single number or 'all', if only a single parameter/error/warning or all
        of them should be listed. 
        """        
        try:
            dict_ = self._dicts[letter]
        except KeyError:
            raise ValueError(f'invalid *letter*: {repr(letter)}')

        # If *numbers* is a single number, convert it into a list.
        if isinstance(numbers, int):
            numbers = [numbers]
        
        self.print(table(dict_, numbers, self._widths[letter]))
        
    def cmd_help(self, value=None):
        """Display a help message.

        *value* should be the name or an alias of the command that
        should be described.
        If no *value* is specified, all commands are listed and
        described.
        """
        if value:
            string = self._helpstring(value)
        else:
            descriptions = []
            for name, _ in self.cmds_and_aliases:
                descriptions.append(self._helpstring(name))
            description_str = '\n\n'.join(descriptions)
            string = ('Valid commands:\n'
                      + textwrap.indent(description_str, self.indent))

        self.print(string)

    def cmd_docs(self):
        """Open TurboCtl documentation in a web browser."""
        docs()

    def _helpstring(self, cmdname):
        """Return a help message describing the usage of *cmdname*."""
        method = self._get_method(cmdname)
        argstr = self._argstring(method)
        aliasstr = self._alias_string(cmdname)
        docstr = inspect.getdoc(method)
        if docstr is None:
            docstr = ''

        return (f'{cmdname} {argstr}\n'
                + textwrap.indent(docstr, self.indent)
                + ('\n\n' + textwrap.indent(aliasstr, self.indent) if aliasstr
                   else '')
                )

    def cmd_exit(self):
        """Exit the UI."""
        self._stop_flag = True

    def cmd_debug(self, value):
        """Activate or deactivate the debug mode.

        Values of '1', 'True' and 'on' activate the debug mode;
        '0', 'False' and 'off' deactivate it.

        In normal operation, TypeErrors and ValueErrors
        raised during the execution of commands are caught to prevent
        users from crashing the program with invalid commands.
        Activating the debug mode disables this error-catching in order
        to make debugging easier.
        """
        # 1 == True and 0 == False.
        if value in [True, 'on']:
            self.debug = True
            self.print('Debug mode activated')

        elif value in [False, 'off']:
            self.debug = False
            self.print('Debug mode deactivated')

        else:
            raise ValueError('invalid value')
            
    def cmd_verbose(self, value):
        """Activate or deactivate the verbose mode.

        Values of '1', 'True' and 'on' activate the verbose mode;
        '0', 'False' and 'off' deactivate it.

        When the verbose mode is on, commands that send telegrams to the pump
        will print all the contents of the telegram and the reply to the
        screen.
        """
        # 1 == True and 0 == False.
        if value in [True, 'on']:
            self.verbose = True
            self.print('Verbose mode activated')

        elif value in [False, 'off']:
            self.verbose = False
            self.print('Verbose mode deactivated')

        else:
            raise ValueError('invalid value')
            
    def input(self, prompt=''):
        """Ask the user for input.

        This method works like the built-in :func:`.input` function,
        but uses :attr:`inputfile` and :attr:`outputfile` instead of
        :data:`sys.stdin` and :data:`sys.stdout`.

        The line editing functionality provided by the :mod:`readline`
        module works with this method only if :attr:`inputfile` is
        :data:`sys.stdin`.
        """
        # `.input` instead of `input` makes Sphinx refer to the
        # built-in input function instead of this method.

        if self.inputfile == sys.stdin:
            return input(prompt)

        self.print(prompt, end='')
        string = self.inputfile.readline()

        if string[-1] == '\n':
            return string[:-1]

        return string

    def print(self, *objects, sep=' ', end='\n'):
        """Print text to the UI.

        This method works like the built-in :func:`.print` function,
        but uses :attr:`inputfile` and :attr:`outputfile` instead of
        :data:`sys.stdin` and :data:`sys.stdout`, and doesn't include
        the *file* and *flush* arguments.
        """
        print(*objects, file=self.outputfile, sep=sep, end=end, flush=True)

    def _run_command(self, line):
        """Parse *line* and run it as a command.

        Returns:
            True if the main loop should be broken, otherwise
            False.
        """
        # Separate the command and its arguments.
        # split() splits at any whitespace character
        # and discards empty strings.
        parts = line.split()
        try:
            cmdname = parts[0]
        except IndexError:
            # Skip empty lines.
            return False
        args = parts[1:]

        # Get the method.
        try:
            method = self._get_method(cmdname)
        except ValueError as error:
            self.print(f'Error: {error}')
            return False

        # Parse the arguments into appropriate Python objects.
        # Arguments that cannot be parsed into any other type are kept
        # as strings.
        for i, arg in enumerate(args):
            try:
                args[i] = ast.literal_eval(arg)
            except (ValueError, SyntaxError):
                pass

        # Call the method.
        try:
            out = method(*args)
            # If the verbose mode is on, print the contents of every telegram
            # that is sent or received.
            # For methods that don't send telegrams, *out* will be None.
            if out and self.verbose:
                self.print('\n' + self._verbose_string(*out))
                
        except (ValueError, TypeError) as error:

            # Suppress errors, unless the debig mode is on.
            if self.debug:
                raise error

            if isinstance(error, ValueError):
                # ValueErrors should contain a meaningful explanation string,
                # so they can be displayed to the user.
                self.print(f'Error: {error}')
            else:
                # TypeErrors will count the number of arguments wrong (since
                # they also count the *self* argument), and other types of
                # errors shouldn't be raised during normal operation, so those
                # aren't displayed to the user.
                self.print('Error: invalid argument type or number of '
                           'arguments')
                
    def _verbose_string(self, query, reply):
        """Return a string displaying the contents of both the query and the
        reply.
        """        
        return (
            'Sent a telegram with the following contents:\n'
            + textwrap.indent(str(query), self.indent) + '\n'
            + '\n'
            + 'Received a telegram with the following contents:\n'
            + textwrap.indent(str(reply), self.indent)
        )

    def _get_method(self, command):
        """Return the method corresponding to *command* (a str)."""

        names = [command] + self._get_aliases(command)
        for name in names:
            try:
                return getattr(self, 'cmd_' + name)
            except AttributeError:
                pass

        raise ValueError(f'invalid command: {command}')

    def _get_aliases(self, command):
        """Return a list of aliases for *command*.

        *command* itself is never part of the list, but if *command*
        is itself an alias, the primary name of the command will be
        included in the list.
        """
        for cmd, aliases in self.cmds_and_aliases:
            names = [cmd] + aliases
            if command in names:
                names.remove(command)
                return names

        raise ValueError(f'invalid command: {command}')

    def _alias_string(self, command):
        """Return a formatted string that lists the aliases of
        "*command*.

        Returns:
            A string with the format 'Aliases: a, b, c',
            or '' if there are no aliases.
        """
        aliases = self._get_aliases(command)
        if aliases:
            return 'Aliases: ' + ', '.join(aliases)
        return ''

    @staticmethod
    def _argstring(method):
        """Return a formatted string that lists the arguments of
        *command*.

        Format: '<arg1> [optional_arg2=default]'
        """
        argspec = inspect.getfullargspec(method)
        args = argspec.args[1:]  # Ignore *self*.
        defaults = argspec.defaults  # Default values for arguments.
        if defaults is None:
            defaults = []

        reversed_args = args[::-1]
        reversed_defaults = defaults[::-1]
        reversed_argstrings = [None] * len(args)

        # The lists are reversed, so that arguments with default
        # values are iterated first.
        # Arguments without default values are iterated afterwards,
        # when *reversed_defaults* has run out of indices.
        for i, arg in enumerate(reversed_args):
            try:
                default = reversed_defaults[i]
                reversed_argstrings[i] = f'[{arg}={default}]'
            except IndexError:
                reversed_argstrings[i] = f'<{arg}>'

        return ' '.join(reversed_argstrings[::-1])



"""Unit tests for the command_parser module."""

import unittest
from collections import namedtuple

from turboctl import Command, CommandParser, UIParseError, UICommandError

Output = namedtuple('Output', 'name, args, kwargs')

cmd_list = [
    Command(names=['test', 't'],
        function='test_method',
        args=[],
        description=''),
]


class DummyUI():
    def test_method(self, *args, **kwargs):
        return Output('test', args, kwargs)


class TestCommandParser(unittest.TestCase):
    
    @classmethod
    def setUpClass(self):
        self.ui = DummyUI()
        self.parser = CommandParser(self.ui, cmd_list)
                
    def test_no_args(self):
        string = 'test'
        out = self.parser.parse(string, False)
        
        self.assertEqual(out.name, 'test')
        self.assertEqual(out.args, ())
        self.assertEqual(out.kwargs, {})
        
    def test_args_and_kwargs(self):
        string = 'test 1 kwarg1=3 2 kwarg2=4'
        out = self.parser.parse(string, False)
        
        self.assertEqual(out.name, 'test')
        self.assertEqual(out.args, (1, 2))
        self.assertEqual(out.kwargs, {'kwarg1': 3, 'kwarg2': 4})
        
    def test_alias(self):
        string = 't 1'
        out = self.parser.parse(string, False)
        
        self.assertEqual(out.name, 'test')
        self.assertEqual(out.args, (1,))
        self.assertEqual(out.kwargs, {})
        
    def test_data_types(self):
        string = "test 1 'str' also_str (1,2) [3,4] dict={5:6} 7,8,9 True"
        out = self.parser.parse(string, False)
        
        self.assertEqual(out.name, 'test')
        self.assertEqual(out.args, 
                         (1, 'str', 'also_str', (1,2), [3,4], (7,8,9), True)
                         )
        self.assertEqual(out.kwargs, {'dict': {5: 6}})
        
    def test_whitespace(self):
        string = "   test     line1\nline2      "
        out = self.parser.parse(string, False)
        
        self.assertEqual(out.name, 'test')
        self.assertEqual(out.args, ('line1\nline2',))
        self.assertEqual(out.kwargs, {})
        
    def test_empty(self):
        string = "   "
        out = self.parser.parse(string, False)
        self.assertFalse(out)
        
    def test_valid_and_invalid_keywords(self):
        invalid_args = ('1arg=1', 'arg*=1', 'ar#g=1', 'arg=arg=1')
        valid_args = ('arg1=1', '_arg_=1', 'ARG=1')
        
        for arg in invalid_args:
            with self.assertRaises(UIParseError):
                self.parser.parse(f'test {arg}', False)
                
        for arg in valid_args:
            out = self.parser.parse(f'test {arg}', False)
        
            self.assertEqual(out.name, 'test')
            self.assertEqual(out.args, ())
            self.assertEqual(set(out.kwargs.values()), set((1,)))
            
    def test_valid_and_invalid_values(self):
        valid_values = ('"', 'aa#aa', "'", '@$¤')
        
        with self.assertRaises(UIParseError):
            self.parser.parse(f'test 1==1', False)

        for v in valid_values:
            out = self.parser.parse(f'test {v}', False)
        
            self.assertEqual(out.name, 'test')
            self.assertEqual(out.args, (v,))
            self.assertEqual(out.kwargs, {})
            
    def test_invalid_commands(self):
        invalid_commands = ('test1', '#¤$', "'")
        
        for cmd in invalid_commands:
            with self.assertRaises(UICommandError):
                self.parser.parse(f'{cmd} 1', False)

    
if __name__ == '__main__':
    unittest.main()


import unittest
import sys
import time
from io import StringIO

from turboctl import (ShellTUI, VirtualPump, Query, Reply, StatusBits, 
                      ControlBits, full_output, control_or_status_output, 
                      hardware_output, parameter_output, 
                      UIArgumentNumberError, UITypeError, UIValueError)      
        
class Base(unittest.TestCase):
    
    def setUp(self):
        self.maxDiff = None
        self.old = sys.stdout
        self.printed = StringIO()
        sys.stdout = self.printed
        self.vp = VirtualPump()
        self.ui = ShellTUI(self.vp.port)
        
    def tearDown(self):
        sys.stdout = self.old
        self.vp.close()
    
    
class TestOnOff(Base):
    
    @classmethod
    def setUpClass(cls):
        cls.q = Query(control_or_status_set={ControlBits.START_STOP, 
                      ControlBits.COMMAND}, frequency=0, temperature=40, 
                      current=10, voltage=20)
        cls.r = Reply(cls.q.data)
        cls.r.control_or_status_set={StatusBits.OPERATION, 
                                     StatusBits.ACCELERATION}
    
    def test_verbosity_1(self):
        self.ui.cmd_verbosity(1)
        self.ui.cmd_onoff()
        string = ('Turning the pump on or off\n')
        self.assertEqual(self.printed.getvalue(), string)
        
    def test_verbosity_2(self):
        self.ui.cmd_verbosity(2)
        self.ui.cmd_onoff()
        string = ('Turning the pump on or off\n' +
                  control_or_status_output(self.r, True) + '\n')
        self.assertEqual(self.printed.getvalue(), string)
        
    def test_verbosity_3(self):
        self.ui.cmd_verbosity(3)
        self.ui.cmd_onoff()
        string = full_output(self.q, self.r) + '\n'
        self.assertEqual(self.printed.getvalue(), string)
        
        
class TestStatus(Base):
    
    @classmethod
    def setUpClass(cls):
        cls.q = Query()
        cls.r = Reply(control_or_status_set={StatusBits.READY})
    
    def test_verbosity_1(self):
        self.ui.cmd_verbosity(1)
        self.ui.cmd_status()
        string = hardware_output(self.r, False) + '\n'
        self.assertEqual(self.printed.getvalue(), string)
        
    def test_verbosity_2(self):
        self.ui.cmd_verbosity(2)
        self.ui.cmd_status()
        string = (control_or_status_output(self.r) + '\n' + 
                  hardware_output(self.r, True) + '\n')
        self.assertEqual(self.printed.getvalue(), string)
        
    def test_verbosity_3(self):
        self.ui.cmd_verbosity(3)
        self.ui.cmd_status()
        string = full_output(self.q, self.r) + '\n'
        self.assertEqual(self.printed.getvalue(), string)
        
        
class TestRead(Base):
    
    @classmethod
    def setUpClass(cls):
        cls.q = Query(parameter_number=1, parameter_mode='read')
        cls.r = Reply(parameter_number=1, parameter_mode='response',
                      parameter_value=180, 
                      control_or_status_set={StatusBits.READY})
    
    def test_verbosity_1(self):
        self.ui.cmd_verbosity(1)
        self.ui.cmd_read(1)
        string = parameter_output(self.r, False) + '\n'
        self.assertEqual(self.printed.getvalue(), string)
        
    def test_verbosity_2(self):
        self.ui.cmd_verbosity(2)
        self.ui.cmd_read(1)
        string = parameter_output(self.r, True) + '\n'
        self.assertEqual(self.printed.getvalue(), string)
        
    def test_verbosity_3(self):
        self.ui.cmd_verbosity(3)
        self.ui.cmd_read(1)
        string = full_output(self.q, self.r) + '\n'
        self.assertEqual(self.printed.getvalue(), string)
        
    def test_index(self):
        self.ui.cmd_verbosity(2)
        self.ui.cmd_read(1, 1)
        r = Reply(self.r.data)
        r.parameter_index = 1
        string = parameter_output(r, True) + '\n'
        self.assertEqual(self.printed.getvalue(), string)
        
    def test_invalid_args(self):
        self.ui.cmd_verbosity(2)
        
        errors = {'1'  : UITypeError,
                  1.0  : UITypeError,
                  -1   : UIValueError,
                  2**16: UIValueError}
        
        for n, e in errors.items():
            with self.assertRaises(e):
                self.ui.cmd_read(n)
                
        for n, e in errors.items():
            with self.assertRaises(e):
                self.ui.cmd_read(1, n)






        
    
    
#    def test_status(self):
#        
#        str_in = 'status'
#        
#        str_out = (
#"""
#Sent a telegram with the following contents:
#    No parameter access
#    No control bits active
#
#Received a telegram with the following contents:
#    No parameter access
#    No status conditions present
#    Stator frequency: 0 Hz
#    Frequency converter temperature: 0 °C
#    Motor current: 0×0.1 A
#    Intermediate circuit voltage: 0×0.1 V
#""")
#        
#        self.ui.run(str_in)
#        self.assertEqual(self.ui.output, str_out)


if __name__ == '__main__':
    unittest.main()


"""This module handles the simulation of pump hardware in a
:class:`~turboctl.virtualpump.virtualpump.VirtualPump`.

..
    Aliases for Sphinx.

.. |Uint| replace:: :class:`~turboctl.telegram.datatypes.Uint`
.. |Sint| replace:: :class:`~turboctl.telegram.datatypes.Sint`
.. |Float| replace:: :class:`~turboctl.telegram.datatypes.Float`
.. |Bin| replace:: :class:`~turboctl.telegram.datatypes.Bin`
.. |ExtendedParameter| replace::
    :class:`~turboctl.virtualpump.parameter_component.ExtendedParameter`
.. |ParameterComponent| replace::
    :class:`~turboctl.virtualpump.parameter_component.ParameterComponent`
.. |ParameterComponent.handle_parameter| replace::
    :meth:`ParameterComponent.handle_parameter()
    <turboctl.virtualpump.parameter_component.ParameterComponent\
.handle_parameter>`
"""

import threading
import time

from turboctl.telegram.codes import StatusBits, ControlBits


class HardwareComponent():
    """This class defines the part of a
    :class:`~turboctl.virtualpump.virtualpump.VirtualPump`
    that handles hardware data and command/status bits.
    
    The following pump properties are simulated:
        
        - The pump can be turned on and off.
        
        - Temperature, current and voltage are 0 when the pump is off, and
          have constant positive values when the pump is on. This is reflected 
          both in the reply telegrams and the values of corresponding
          parameters.
          
        - When the pump is on, the pump frequency moves towards the setpoint at
          a constant rate. When the pump is off, the frequency moves towards 0
          at the same rate.
          
        - The frequency setpoint is always read from the corresponding
          parameter, and cannot currently be set with the frequency field of
          a telegram.
          
        - Currently this class only recognizes the ``COMMAND`` and ``ON``
          control bits (see :class:`~turboctl.telegram.codes.ControlBits`);
          all others are ignored.
          
        - The following status bits
          (see :class:`~turboctl.telegram.codes.StatusBits`)
          are applied to the reply telegram:
            
              - ``OPERATION'``: When the pump is on.
              - ``READY``: When the pump is off.
              - ``TURNING``: When the pump frequency is not 0.
              - ``ACCELERATION``: When the pump acceleration is above 0.
              - ``DECELERATION``: When the pump acceleration is below 0.
              - ``PARAM_CHANNEL``: Always.
              - ``PROCESS_CHANNEL``: When the ``COMMAND`` control bit is
                supplied.

    Attributes:
        variables:
            A :class:`Variables` instance that contains the parameters that
            can be modified by this object.
            
        step:
            The timestep of iteration; the parallel thread that updates
            :attr:`frequency` waits this amount of seconds between iterations.
            
        abs_acceleration:
            How fast the pump accelerates or decelerates, in Hz / second.
            
        frequency:
            The exact frequency of the pump as a :class:`float`.

            This is needed to correctly simulate gradual changes in the
            frequency, because the frequency parameter only saves integer
            values.
            
        is_on:
            A :class:`bool` flag to keep track of whether the pump is on
            or off.

        shutoff_time:
            The automatic shutoff interval of the pump in seconds.

            The pump automatically turns off after :attr:`shutoff_time`
            seconds have elapsed since the last :meth:`on` command.
            The real pump has a shutoff time of 10 seconds, but the value of
            this attribute can be lowered for testing purposes. 

        on_time:
            How long (in seconds) has elapsed since the last :meth:`on`
            command.
    """

    TEMPERATURE = 30
    """The pump reports this constant temperature when it's on.
    The unit is °C."""
    
    CURRENT = 10
    """The pump reports this constant temperature when it's on.
    The unit is 0.1 A.
    """
    
    VOLTAGE = 24
    """The pump reports this constant temperature when it's on.
    The unit is 0.1 V.
    """
    
    def __init__(self, parameters, lock):
        """Initialize a new :class:`HardwareComponent`.
        
        Args:
            parameters: The same :class:`dict` of |ExtendedParameter|
                objects as used by |ParameterComponent|.
                This is needed because some hardware components affect the
                values of parameters and vice versa.
                Note that these need to be the actual parameters of the pump;
                dummy values cannot be used even for testing purposes, since
                all hardware-related parameters (see :class:`HWParameters` for
                a list of those) need to be present for this class to function.
                
            lock: A :class:`threading.Lock` object that can be used to
                temporarily freeze the parallel thread which updates
                :attr:`frequency`.
                The purpose of this is to prevent race conditions between that
                thread and the :meth:`handle_hardware` and
                |ParameterComponent.handle_parameter| methods, since they all
                can access and modify the value of the frequency parameter.
        """
        self.parameters = parameters
        self.variables = Variables(parameters)
        
        self.step = 0.1
        self.abs_acceleration = 100
        self.frequency = 0.0
        
        # The pump starts in an off state.
        self.is_on = False
        self.shutoff_time = 10
        self.on_time = 0
        # This sets some parameters to the off state.
        self.off()
                
        # Start a parallel thread for continuously updating the pump 
        # frequency.
        thread = threading.Thread(target=self._run, args=[lock])
        self._stop_flag = threading.Event()
        thread.daemon = True
        thread.start()
                                
    def handle_hardware(self, query, reply):
        """Write hardware data to *reply.*
        
        Apply the hardware-related commands specified by *query*, 
        change pump attributes accordingly, and write hardware data and
        status bits to *reply*.
        
        Args:
            query (:class:`~turboctl.telegram.telegram.TelegramReader`):
                The telegram sent to the pump.
                
            reply(:class:`~turboctl.telegram.telegram.TelegramBuilder`):
                This will be used to build the telegram sent from the pump.                
        """
        self._handle_control_bits(query)
        self._handle_status_bits(query, reply)
        
        (reply.set_frequency(self.variables.frequency)
              .set_temperature(self.variables.temperature)
              .set_current(self.variables.current)
              .set_voltage(self.variables.voltage))
        
    def _handle_control_bits(self, query):
        """Apply the effects of control bits in *query*.

        Currently only the COMMAND and ON control bits are
        recognized; all others are ignored.
        """
        on_command = (ControlBits.COMMAND in query.flag_bits and
                      ControlBits.ON in query.flag_bits)

        off_command = (ControlBits.COMMAND in query.flag_bits and
                       ControlBits.ON not in query.flag_bits)

        if on_command:
            self.on()

        if off_command:
            self.off()

    def _handle_status_bits(self, query, reply):
        """Write appropriate status bits to *reply*.
        
        The following status bits are supported:
            OPERATION
            READY
            TURNING
            ACCELERATION
            DECELERATION
            PARAM_CHANNEL
            PROCESS_CHANNEL
        """
        # TODO: find out if this is correct by testing the real pump.
        
        bitlist = [StatusBits.PARAM_CHANNEL]
        
        if ControlBits.COMMAND in query.flag_bits:
            bitlist.append(StatusBits.PROCESS_CHANNEL)
          
        if self.is_on:
            bitlist.append(StatusBits.OPERATION)
        else:
            bitlist.append(StatusBits.READY)
            
        if self.variables.frequency:
            bitlist.append(StatusBits.TURNING)
            
        if self._acceleration > 0:
            bitlist.append(StatusBits.ACCELERATION)
            
        if self._acceleration < 0:
            bitlist.append(StatusBits.DECELERATION)
            
        reply.set_flag_bits(bitlist)
            
    def stop(self):
        """Order the parallel thread to stop."""
        self._stop_flag.set()        
        
    def on(self):
        """Turn the pump on and update parameters accordingly."""
        self.is_on = True
        self.on_time = time.time()
        self.variables.temperature = self.TEMPERATURE
        self.variables.current = self.CURRENT
        self.variables.voltage = self.VOLTAGE
        
    def off(self):
        """Turn the pump off and update parameters accordingly."""
        self.is_on = False
        self.variables.temperature = 0
        self.variables.current = 0
        self.variables.voltage = 0
     
    @property
    def _acceleration(self):
        """Return the current acceleration of the frequency in Hz/s.
        
        Deceleration is represented by a negative acceleration.
        """
        if self.frequency < self._frequency_goal:
            return self.abs_acceleration
        elif self.frequency > self._frequency_goal:
            return -self.abs_acceleration
        else:
            return 0
        
    @property
    def _frequency_goal(self):
        """Return the frequency setpoint if the pump is on, or 0 if 
        the pump is off.
        """
        return self.variables.frequency_setpoint if self.is_on else 0
    
    def _change_frequency(self):
        """Enact the frequency change during a single time-step."""
        change = self.step * self._acceleration
        difference = self._frequency_goal - self.frequency
        
        # Make sure the change never takes the frequency to the other side of
        # the frequency goal.
        if abs(change) > abs(difference):
            change = difference
            
        self.frequency += change
        self.variables.frequency = int(self.frequency)
                
    def _run(self, lock):
        """Run the parallel thread by continuously updating the 
        pump status.
        """
        while not self._stop_flag.is_set():
            time.sleep(self.step)
            lock.acquire()

            # Turn the pump off if a long enough time has elapsed since the
            # last "pump on" command.  
            timed_shutoff = time.time() - self.on_time > self.shutoff_time
            if self.is_on and timed_shutoff:
                self.off()

            # Update the frequency.
            self._change_frequency()
            lock.release()


class HWParameters():
    """A collection of parameters related to pump hardware.
    
    Each of the attributes of this class is an |ExtendedParameter|
    object. This class gives these parameters descriptive names and makes it
    possible to access them without having to find out the parameter numbers.
    
    Note that :class:`HardwareComponent` currently cannot modify most of these
    parameters, but they are included in case :class:`HardwareComponent` is
    expanded in the future.
    
    Attributes:
        
        frequency:
            Parameter 3. Current rotor frequency in Hz.
            This value is included in every reply telegram.
            
            Type: 16-bit |Uint|.
            
        voltage:
            Parameter 4. Current intermediate circuit voltage in 0.1 V.
            This value is included in every reply telegram.
            
            Type: 16-bit |Uint|.
            
        current:
            Parameter 5. Current motor current in 0.1 A.
            This value is included in every reply telegram.
            
            Type: 16-bit |Uint|.
            
        motor_power:
            Parameter 6. Current motor input power in 0.1 W.
            
            Type: 16-bit |Uint|.
            
        motor_temperature:
            Parameter 7. Current motor temperature in 0.1 °C.
            
            Type: 16-bit |Sint|.
            
        save_data:
            Parameter 8. Writing any value to this parameter saves all
            changed parameter values into nonvolatile memory. Otherwise all
            changes are lost when power to the pump is cut off.
            
            Type: 16-bit |Sint|.
            
        temperature:
            Parameter 11. Current frequency converter temperature in °C.
            This value is included in every reply telegram.
            
            Type: 16-bit |Sint|.
            
        frequency_setpoint:
            Parameter 24. Setpoint for the rotor frequency in Hz.
            The pump tries to keep the frequency at this number.
            
            Type: 16-bit |Uint|.
                
        error_counter:
            Parameter 40. Counts the total number of error conditions that
            have occurred to the pump.
            
            Type: 16-bit |Uint|.
            
        overload_error_counter:
            Parameter 41. Counts the number of error conditions caused by an
            overload. An overload error is represented by the error code 106.
            
            Type: 16-bit |Uint|.
            
        power_failure_error_counter: 
            Parameter 43. Counts the number of error conditions caused by a
            disturbance in power supply. A power supply error is reprsented
            by the error code 603.
            
            Type: 16-bit |Uint|.
            
        error_list:
            Parameter 171. This parameter has 254 indices, each of which
            stores an error code. The most recent error is located at index 0
            and the oldest at index 253.
            
            Type: 16-bit |Uint|, 254 indices.
            
        error_frequency_list:
            Parameter 174. The indices of this parameter contain the frequency
            of the pump at the time of the corresponding error in
            :attr:`error_list`.
            
            Type: 16-bit |Uint|, 254 indices.
            
        error_hour_list:
            Parameter 176. This parameter works analogously to
            :attr:`error_frequency_list`, but contains the number of
            operational hours of the pump instead of the frequency. 
            
            Type: 32-bit |Sint|, 254 indices.
            
        operating_hours: 
            Parameter 184. Counts the total number of operational hours for
            the pump.
            
            Type: 32-bit |Sint|, 254 indices.
            
        warning_list:
            Parameter 227. Works analogously to :attr:`error_list`, but lists
            warnings instead of errors.
            
            Type: 16-bit |Uint|, 254 indices.
    """
    # Parameter 6 is listed as "drive power" in the manual, but this is
    # probably more or less the same as the motor.
    
    def __init__(self, parameters):
        """Initializer.
        
        Args:
            parameters: A :class:`dict` of |ExtendedParameter| objects,
                which needs to include at least all the parameters which are
                attributes of this class.
        """
        self.frequency = parameters[3]
        self.voltage = parameters[4]
        self.current = parameters[5]
        self.motor_power = parameters[6]
        self.motor_temperature = parameters[7]
        self.save_data = parameters[8]
        self.temperature = parameters[11]
        self.frequency_setpoint = parameters[24]
        self.error_counter = parameters[40]
        self.overload_error_counter = parameters[41]
        self.power_failure_error_counter = parameters[43]
        self.error_list = parameters[171]
        self.error_frequency_list = parameters[174]
        self.error_hour_list = parameters[176]
        self.operating_hours = parameters[184]
        self.warnings = parameters[227]


class Variables():
    """This class encapsulates :class:`HWParameters` in order to allow the
    simpler syntax
    ::
        
        variables.parameter_name = value
        
    instead of
    ::
        
        parameter = hwparameters.parameter_name
        parameter.value[0] = parameter.datatype(value, parameter.bits)
        
    Attributes:
        parameters: The encapsulated :class:`HWParameters` object.
    """
    
    def __init__(self, parameters):
        """Initialize a new :class:`Variables` instance.
        
        Args:
            parameters: Passed on to :meth:`HWParameters.__init__`.
        """
        self.parameters = HWParameters(parameters)
        
    def __setattr__(self, name, value):
        """Set the attribute *name* to *value*.
        
        The :attr:`parameters` attribute is handled as a special case and
        is accessed normally.
        For all other attributes, the value of the parameter in
        :attr:`parameters` with the name *name* is set to *value*.
        
        *value* should be given as an instance of a built-in Python type, and
        will automatically be converted to a
        :class:`~turboctl.telegram.datatypes.Data` subclass instance.
        However, note that the type of *value* should match the type of the
        parameter; e.g. a :class:`float` can be converted into a |FLoat| but
        not into an |Uint|.
        """
        if name == 'parameters':
            self.__dict__[name] = value
            return
        
        parameter = getattr(self.parameters, name)
        # Even the values of unindexed parameters are lists.
        parameter.value[0] = parameter.datatype(value, parameter.bits)
        
    def __getattr__(self, name):
        """Return the the value of the attribute *name*.
        
        This method returns the value of the parameter in :attr:`parameters`
        with the name *name*.
        The value is automatically converted to a built-in type.
        
        Since ``__getattr__`` (unlike ``__getattribute__``) is only called if
        the attribute cannot be found through normal routes, there is no need
        to make :attr:`parameters` into a special case.
        """        
        parameter = getattr(self.parameters, name)
        return parameter.value[0].value



"""This module defines an API of functions which can be used to send commands
to the pump.

All functions in this module share the following common arguments:
    
    *connection*:
        This is a :class:`serial.Serial` instance, which is used to send the
        command.
        
    *pump_on*:
        If this evaluates to ``True``, control bits telling the pump to turn or
        stay on are added to the telegram. Otherwise receiving the telegram
        will cause the pump to turn off if it is on.  

If a command cannot be sent due to an error in the connection, a
:class:`serial.SerialException` will be raised.

The functions return both the query sent to the pump and the reply received
back as :class:`~turboctl.telegram.telegram.TelegramReader` instances.
"""

from turboctl.telegram.codes import ControlBits
from turboctl.telegram.telegram import (Telegram, TelegramBuilder, 
                                        TelegramReader)


_PUMP_ON_BITS = [ControlBits.COMMAND, ControlBits.ON]
_PUMP_OFF_BITS = [ControlBits.COMMAND]

           
def send(connection, telegram):
    """Send *telegram* to the pump.
    
    Args:
        telegram: A :class:`~turboctl.telegram.telegram.Telegram` instance.
    """
    connection.write(bytes(telegram))
    reply_bytes = connection.read(Telegram.LENGTH)
    reply = TelegramBuilder().from_bytes(reply_bytes).build()
    return TelegramReader(telegram, 'query'), TelegramReader(reply, 'reply')


def status(connection, pump_on=None):
    """Request pump status.
    
    This function sends an empty telegram to the pump, which causes it to send
    back a reply containing some data about the status of the pump.
    
    This can also be used for turning the pump on or off by setting *pump_on*
    to ``True`` or ``False``.
    """
    builder = TelegramBuilder()
    match pump_on:
        case False:
            builder.set_flag_bits(_PUMP_OFF_BITS)
        case True:
            builder.set_flag_bits(_PUMP_ON_BITS)
    query = builder.build()
    return send(connection, query)


def reset_error(connection):
    """Reset the error status of the pump.
    
    This function sends a "reset error" command to the pump.
    """
    builder = TelegramBuilder()
    clear_error = [ControlBits.COMMAND, ControlBits.RESET_ERROR]
    builder.set_flag_bits(clear_error)
    query = builder.build()
    return send(connection, query)


def _access_parameter(connection, mode, number, value, index, pump_on):
    """This auxiliary function provides functionality for both reading and
    writing parameter values, since the processes are very similar.
    """
    builder = (TelegramBuilder()
        .set_parameter_mode(mode)
        .set_parameter_number(number)
        .set_parameter_index(index)
        .set_parameter_value(value)
    )
    
    if pump_on:
        builder.set_flag_bits(_PUMP_ON_BITS)
    
    query = builder.build()
    return send(connection, query)

    
def read_parameter(connection, number, index=0, pump_on=True):
    """Read the value of an index of a parameter.
    
    Args:
        number
            The number of the parameter.
            
        index
            The index of the parameter (0 for unindexed parameters).
            
    Raises:
        ValueError:
            If *number* or *index* have invalid values.
    """
    return _access_parameter(connection, 'read', number, 0, index, pump_on)


def write_parameter(connection, number, value, index=0, pump_on=True):
    """Write a value to an index of a parameter.
    
    Args:
        number:
            The number of the parameter.
                        
        value:
            The value to be written.
            
        index:
            The index of the parameter (0 for unindexed parameters).
            
    Raises:
        ValueError:
            If *number* or *index* have invalid values.
    """
    return _access_parameter(
        connection, 'write', number, value, index, pump_on)


"""This module defines classes representing different data types used
in telegrams.

..
    Aliases for Sphinx.

.. |value| replace:: :attr:`value <Data.value>`
.. |bits| replace:: :attr:`bits <Data.bits>`
.. |n_bytes| replace:: :attr:`n_bytes <Data.n_bytes>`
.. |BYTESIZE| replace:: :const:`BYTESIZE`
"""
from functools import singledispatchmethod 
import math
import re
import struct
from typing import Optional


BYTESIZE = 8
"""The number of bits in a byte."""


def maxuint(bits):
    """Return the largest unsigned integer that can be expressed with
    *bits* bits.
    
    If *bits* is ``0``, ``0`` is returned.
    
    Args:
        bits:
            A non-negative :class:`int`.
    """
    return 2**bits - 1
    

def maxsint(bits):
    """Return the largest (i.e. most positive) signed integer that can
    be expressed with *bits* bits.
    
    If *bits* is ``0``, ``0`` is returned.
    
    Args:
        bits:
            A non-negative :class:`int`.
    """
    if bits == 0:
        return 0
    
    return 2**(bits - 1) - 1


def minsint(bits):
    """Return the smallest (i.e. most negative) signed integer that can
    be expressed with *bits* bits.
    
    If *bits* is ``0``, ``0`` is returned.
    
    Args:
        bits:
            A non-negative :class:`int`.
    """
    if bits == 0:
        return 0
    
    return -2**(bits - 1)


class Data():
    """A superclass for all telegram data types.
    
    This class is not meant to be initialized, since it lacks an
    ``__init__`` method.
    """
    
    @property
    def value(self):
        """The value represented by this data object.
        This is a read-only property.
        """
        return self._value
    
    @property   
    def bits(self) -> int:
        """How many bits of data this object represents.
        This is a read-only property.
        """
        return self._bits 
        
    @property
    def n_bytes(self) -> int:
        """How many bytes are needed to store the data in the object;
        equal to |bits| divided by |BYTESIZE| and rounded up.
        """
        return math.ceil(self.bits / BYTESIZE)

    def __add__(self, other):
        """Return *self* + *other*.
        
        Appends the binary data in *other* to the end of *self* and
        returns an object of the same class as *self* with the
        combined binary data.
        
        Example: ``Bin('010') + Bin('001') = Bin('010001')``
        """
        class_ = type(self)
        string1 = Bin(self).value
        string2 = Bin(other).value
        return class_(Bin(string1 + string2))

    def __eq__(self, other):
        """Return ``self == other``.
        
        Returns ``True``, if *self* and *other* have the same type,
        |value| and |bits|, otherwise ``False``.
        Two objects with a value of ``NaN`` are equal, if their types
        and bits match.
        """
        if type(self) != type(other):
            return False
        
        if self.bits != other.bits:
            return False
        
        equal_values = self.value == other.value

        try:
            both_nans = math.isnan(self.value) and math.isnan(other.value)
        except TypeError:
            both_nans = False
        
        return equal_values or both_nans
        
    def __repr__(self):
        """Return ``repr(self)``.
        
        The returned string uses the format
        ``'ClassName(<value>, bits=<bits>).'``
        """
        class_ = type(self).__name__
        value = repr(self.value)
        bits = self.bits
        return f'{class_}({value}, bits={bits})'
                 
    def __getitem__(self, key):
        """Return ``self[key]``.
        
        This method uses *key* to get an index or a slice
        of the binary data in *self*,
        and returns an object of the same class as *self* containing
        the binary data in the index or slice.
        
        Example:
            >>> Bin('010001')[0:3]
            Bin('010', bits=3)
        """
        class_ = type(self)
        string = Bin(self).value[key]
        return class_(Bin(string))
    
    
class Uint(Data):
    """A data type for unsigned integers."""
            
    @singledispatchmethod
    def __init__(self, value, bits: int=BYTESIZE):
        """Initialize a new :class:`Uint`.
        
        This method may be called with any of the following signatures:
            
        ::
            
            Sint(value: int, bits: int=8)
            Sint(value: Data)
            Sint(value: bytes)
            
        If *value* is an :class:`int`, |value| and |bits| will be set
        to the values given as arguments.
        
        If *value* is an instance of a subclass of :class:`Data`,
        the initialized object will represent exactly the same binary
        data as *value* and have the same |bits|.
        
        If *value* is a :class:`bytes` object, the initialized
        object will represent the same binary data as *value*.  
        Because :class:`bytes` objects represent sequences of full
        bytes, |bits| will be |BYTESIZE| multiplied by ``len(value)``.
        
        Raises:
            TypeError or ValueError:
                If *value* or *bits* have invalid types or values.
            TypeError:
                If the *bits* argument is supplied when *value* is
                not an :class:`int`. 
        """
        raise TypeError(f'invalid type for *value*: {type(value)}')
        
    @__init__.register
    def _from_int(self, value: int, bits: int=BYTESIZE):
        _check_uint(value, bits)
        self._value = value
        self._bits = bits        

    @__init__.register
    def _from_data(self, value: Data):
        bytes_ = bytes(value)
        i = int.from_bytes(bytes_, 'big')
        bits = value.bits
        self._from_int(i, bits)

    @__init__.register
    def _from_bytes(self, value: bytes):
        i = int.from_bytes(value, 'big')
        bits = len(value) * BYTESIZE
        self._from_int(i, bits)
        
    def __bytes__(self):
        """Return ``bytes(self)``.
        
        Returns the binary data represented by this object as a
        :class:`bytes` object with a length of |n_bytes|. 
        """
        return self.value.to_bytes(self.n_bytes, 'big')
                            
    
class Sint(Data):
    """A data type for signed integers formed with the
    `two's complement
    <https://en.wikipedia.org/wiki/Two%27s_complement>`_ method.
    """

    @singledispatchmethod
    def __init__(self, value, bits: int=BYTESIZE):
        """Initialize a new :class:`Sint`.
        
        This method works like :meth:`Uint.__init__`,
        except that the range of valid :class:`int` values it
        accepts is different.
        """
        raise TypeError(f'invalid type for *value*: {type(value)}')
        
    @__init__.register
    def _from_int(self, value: int, bits: int=BYTESIZE):
        _check_sint(value, bits)
        self._value = value
        self._bits = bits        

    @__init__.register
    def _from_data(self, value: Data):
        bytes_ = bytes(value)
        # int.from_bytes(bytes_, 'big', signed=True)
        # only works if bits == BYTESIZE.
        i = int.from_bytes(bytes_, 'big')
        bits = value.bits
        
        # Convert i to a signed integer.
        if i > maxsint(bits):
            i -= 2**bits
        
        self._from_int(i, bits)

    @__init__.register
    def _from_bytes(self, value: bytes):
        i = int.from_bytes(value, 'big', signed=True)
        bits = len(value) * BYTESIZE
        self._from_int(i, bits)
            
    def __bytes__(self):
        """Return ``bytes(self)``.
        
        See :meth:`Uint.__bytes__` for details.
        """
        # Convert self.value to an unsigned integer.
        i = self.value
        if i < 0:
            i += 2**self.bits

        # self.value.to_bytes(self.n_bytes, 'big', signed=True)
        # only works if self.bits == BYTESIZE.
        return i.to_bytes(self.n_bytes, 'big')


class Float(Data):
    """A data type for 
    `IEEE 754 single-precision
    <https://en.wikipedia.org/wiki/
    Single-precision_floating-point_format
    #IEEE_754_single-precision_binary_floating-point_format:
    _binary32>`_  floating point numbers.
    """
    
    @singledispatchmethod
    def __init__(self, value, bits=4*BYTESIZE):
        """Initialize a new :class:`Float`.
        
        This method may be called with any of the following signatures:
            
        ::
            
            Float(value: float, bits: int=32)
            Float(value: int, bits: int=32)
            Float(value: Data)
            Float(value: bytes)
            
        The method works like :meth:`Uint.__init__`,
        with the following exceptions:
        
        - If *value* is given as a number, it can be either a :class:`float`
          or an :class:`int`. :class:`int` values are automatically converted
          into a corresponding :class:`float` value.
        - *bits* must always be ``32``; it exists as an argument
          only to give all ``__init__`` methods of :class:`Data`
          subclasses a similar signature.
          Likewise, if *value* is a :class:`Data` or a
          :class:`bytes` object, it must contain exactly 32
          bits of data.
        
        Raises:
            ValueError:
                If *bits* is not ``32``, or *value* can't be expressed in
                32 bits of data.
        """
        raise TypeError(f'invalid type for *value*: {type(value)}')
        
    @__init__.register
    def _from_float(self, value: float, bits: int=4*BYTESIZE):
        _check_float(value, bits)
        # Values that are too close to 0 to be expressed as a float are
        # rounded to 0.
        self._value = struct.unpack('>f', struct.pack('>f', value))[0]
        self._bits = bits
        
    #  This is needed so that that datatype(0) works for all datatypes.
    @__init__.register
    def _from_int(self, value: int, bits: int=4*BYTESIZE):
        x = float(value)
        self._from_float(x, bits)

    @__init__.register
    def _from_data(self, value: Data):
        bytes_ = bytes(value)
        x = struct.unpack('>f', bytes_)[0]
        bits = value.bits
        self._from_float(x, bits)

    @__init__.register
    def _from_bytes(self, value: bytes):
        x = struct.unpack('>f', value)[0]
        bits = len(value) * BYTESIZE
        self._from_float(x, bits)
            
    def __bytes__(self):
        """Return ``bytes(self)``.
        
        See :meth:`Uint.__bytes__` for details.
        """
        return struct.pack('>f', self.value)
        
    def __add__(self, other):
        """Return *self* + *other*.
        
        This method works like :meth:`Data.__add__`, except that the
        returned object is a :class:`Bin` instead of a :class:`Float`,
        because :class:`Float` objects cannot have more than 32 bits.
        """
        string1 = Bin(self).value
        string2 = Bin(other).value
        return Bin(string1 + string2)
    
    
    def __getitem__(self, key):
        """Return ``self[key]``.
        
        This method works like :meth:`Data.__getitem__`,
        except that the returned object is a :class:`Bin` instead of a
        :class:`Float`,
        because :class:`Float` objects cannot have less than 32 bits.
        """
        return Bin(self)[key]


class Bin(Data):
    
    @singledispatchmethod
    def __init__(self, value, bits: Optional[int]=None):
        """Initialize a new :class:`Bin`.
        
        This method may be called with any of the following signatures:
            
        ::
            
            Bin(value: str, bits: Optional[int]=None)
            Bin(value: int, bits: int=8)
            Bin(value: Data)
            Bin(value: bytes)
        
        The method works like :meth:`Uint.__init__`, with the following
        exceptions:
            
        - If *value* is specified directly, it can be either a :class:`str`
          or an :class:`int`. If it is a :class:`str`, it must be composed
          solely of the characters ``'1'`` and ``'0'``, or be an empty string.
          If it is an :class:`int`, it must be a valid *bits* bit unsigned
          integer, which will be automatically converted into its binary
          representation.
        
        - If *value* is a :class:`str` and *bits* is ``None``, *bits* will be
          set to the length of *value*.  If *value* is a :class:`str` and
          *bits* is not ``None``, *value* is padded with zeroes to a length of
          *bits*. Giving *bits* a value that is smaller than the length of
          *value* will raise a :class:`ValueError`.
        """
        raise TypeError(f'invalid type for *value*: {type(value)}')
        
    @__init__.register
    def _from_str(self, value: str, bits: Optional[int]=None):
        _check_bin(value, bits)
        self._value = value
        if bits == None:
            bits = len(self.value)
        self._bits = bits
        
    #  This is needed so that that datatype(0) works for all datatypes.
    @__init__.register
    def _from_int(self, value: int, bits: int=BYTESIZE):
        s = _bin_str(value, bits)
        self._from_str(s, bits)
                
    @__init__.register
    def _from_data(self, value: Data):
        bytes_ = bytes(value)
        i = int.from_bytes(bytes_, 'big')
        bits = value.bits
        s = _bin_str(i, bits)
        self._from_str(s, bits)

    @__init__.register
    def _from_bytes(self, value: bytes):
        i = int.from_bytes(value, 'big')
        bits = len(value) * BYTESIZE
        s = _bin_str(i, bits)
        self._from_str(s, bits)
                    
    def __bytes__(self):
        """Return ``bytes(self)``.
        
        See :meth:`Uint.__bytes__` for details.
        """
        if not self.value:
            return b''
        
        return int(self.value, 2).to_bytes(self.n_bytes, 'big')


def _check_uint(value, bits=None):
    """Make sure *value* is a valid *bits* bit unsigned integer.
    
    If *bits* is None, *value* can have any number of bits.
    
    Raises:
        TypeError or ValueError:
            If *value* or *bits* have invalid values or types.
    """
    if not isinstance(value, int):
        raise TypeError(f'value is not an int: {value}')
        
    if value < 0:
        raise ValueError(f'value is negative: {value}')
    
    if bits is not None:
        _check_uint(bits)
        
        if value > maxuint(bits):
            raise ValueError(f'value is too large: {value}')
    
    
def _check_sint(value, bits=None):
    """Like _check_uint, but for signed integers."""
    if not isinstance(value, int):
        raise TypeError(f'value is not an int: {value}')
        
    if bits is not None:
        _check_uint(bits)
        
        if value < minsint(bits) or value > maxsint(bits):
            raise ValueError(f'value is too large or too small: {value}')


def _check_float(value, bits=4*BYTESIZE):
    """Like _check_uint, but for floats.
    
    Raises:
        TypeError or ValueError:
            If *bits* is not 32 (in addition to the exceptions raised by
            _check_uint).
    """
    if not isinstance(value, float):
        raise TypeError(f'value is not a float: {value}')
        
    _check_uint(bits)
        
    if bits != 4*BYTESIZE:
        raise ValueError(f'*bits* should be 32, not {bits}')
        
    try:
        struct.pack('>f', value)
    except OverflowError:
        raise ValueError(f'value is too large or too small: {value}')
    
    
def _check_bin(value, bits=None):
    """Like _check_uint, but for binary strings.
    
    A binary string is '' or a string of '1's and '0's.
    If *bits* is None, it is set to len(value).
    
    Raises:
        ValueError:
            If len(value) != bits
            (in addition to the exceptions raised by '_check_uint').
    """
    # Print 'repr(value)' instead of 'value',
    # since *value* is probably a string.
    if not isinstance(value, str):
        raise TypeError(f'value is not a str: {repr(value)}')
        
    if bits is None:
        bits = len(value)
        
    _check_uint(bits)
        
    if bits != len(value):
        raise ValueError(
            f'bits != len(value); bits={bits}, value={repr(value)}')

    regex=f'\\A[01]{{{bits}}}\\Z'
    if not re.match(regex, value):
        raise ValueError(
            f'{repr(value)} is not a {bits} bit binary string')


def _bin_str(i, bits):
    """Return a *bits* bit binary representation of i.
    
    *i* must be a valid *bits* bit unsigned integer.
    
    E.g. _bin_str(123, 8) returns '01111011'.
        
    Raises:
        TypeError or ValueError:
            If *value* or *bits* have invalid values or types.
    """
    _check_uint(i, bits)

    # _bin_str(0, 0) should return '', which is easiest to handle as a special
    # case, since the algorithm below cannot produce empty strings.
    if (bits == 0):
        return ''

    return (
        bin(i) # Returns something like '0b101'.
        [2:] # Remove '0b'.
        .zfill(bits) # Pad with zeroes to a length of *bits*.
    )



"""This module defines enums for the different codes and numbers used 
in TURBOVAC telegrams.

Members of enums can be accessed with any of the following syntaxes: 
    
    >>> member = EnumName.MEMBER_NAME
    >>> member = EnumName['MEMBER_NAME']
    >>> member = EnumName(member_value)
    >>> member = EnumName(member)
"""

import enum as e


### Parameter access and response codes ###

class ParameterCode(e.Enum):
    """A superclass for parameter access and response codes.
    
    A telegram that tells the pump to accesses a parameter must contain an 
    access code, which depends on the type of the parameter and the access 
    mode.
    The reply telegram sent by the pump then contains a response code which
    also depends on the parameter and the access mode, and also tells whether
    the access was successful.
    
    This enum doesn't have any members, since it's only meant to 
    be subclassed. Members of subclasses have the following fields:
                
        **value**
            The code (a 4-character :class:`str`).
        
        **mode**
            A string that groups the codes together by function 
            (e.g. ``read`` for all read modes).
        
        **indexed**
            ``True`` if this mode can be only used for indexed parameters, 
            ``False`` if it can only be used for unindexed parameters, 
            and ``...`` if it can be used for both.
        
        **bits**
            ``16`` or ``32`` if the mode can only be used for 16 or 32 bit 
            parameters, and ``...`` if it can be used for both.
        
        **description** 
            A verbal description of the meaning of the code.
    """
        
    def __new__(cls, value, mode, indexed, bits, description):
        # __new__ is defined instead of __init__, because setting 
        # _value_ in __init__ prevents the syntax 
        # 'member = EnumName(value_of_member)' from working.
        # This is a comment instead of a docstring to prevent this from 
        # showing up in the docs.
        obj = object.__new__(cls)
        obj._value_ = value
        obj.mode = mode
        obj.indexed = indexed
        obj.bits = bits
        obj.description = description
        return obj
    
    # This doesn't show up in the docs for some reason, but doesn't
    # necessarily need to.
    def __repr__(self):
        """repr(self) returns '<ParameterAccess.XYZ: 1234>' by default.
        It must be overridden so that the syntax 
        "copy = eval(repr(x))" 
        works.
        """
        return str(self)
    

class ParameterAccess(ParameterCode):
    """Different parameter access modes.
    
    This enum contains the following members:
    
    =========== ========== =========== ========= ======= ========================
    Member name Value      Mode        Indexed   Bits    Description
    =========== ========== =========== ========= ======= ========================
    ``NONE``    ``'0000'`` ``'none'``  ``...``   ``...`` ``'No access'``       
    ``R``       ``'0001'`` ``'read'``  ``False`` ``...`` ``'Read a value'``        
    ``W16``     ``'0010'`` ``'write'`` ``False`` ``16``  ``'Write a 16 bit value'``           
    ``W32``     ``'0011'`` ``'write'`` ``False`` ``32``  ``'Write a 32 bit value'``           
    ``RF``      ``'0110'`` ``'read'``  ``True``  ``...`` ``'Read a field value'``                    
    ``W16F``    ``'0111'`` ``'write'`` ``True``  ``16``  ``'Write a 16 bit field value'``                   
    ``W32F``    ``'1000'`` ``'write'`` ``True``  ``32``  ``'Write a 32 bit field value'``                  
    =========== ========== =========== ========= ======= ========================
    """
    # Sphinx doesn't work well with enums, so the members have to be documented
    # by hand.
    # These tables are allowed to exceed the maximum line length, since
    # breaking the lines would seriously hurt readability. 

#   Name     Value   Mode    Indexed  Bits  Description
    NONE = ('0000', 'none',  ...,     ..., 'No access')
    R    = ('0001', 'read',  False,   ..., 'Read a value')
    W16  = ('0010', 'write', False,    16, 'Write a 16 bit value')
    W32  = ('0011', 'write', False,    32, 'Write a 32 bit value')
    RF   = ('0110', 'read',  True,    ..., 'Read a field value')
    W16F = ('0111', 'write', True,     16, 'Write a 16 bit field value')
    W32F = ('1000', 'write', True,     32, 'Write a 32 bit field value')            


class ParameterResponse(ParameterCode):
    """Different parameter response modes.
    
    This enum contains the following members:
    
    ============ ========== ============== ========= ======= =============================
    Member name  Value      Mode           Indexed   Bits    Description
    ============ ========== ============== ========= ======= =============================
    ``NONE``     ``'0000'`` ``'none'``     ``...``   ``...`` ``'No response'``       
    ``S16``      ``'0001'`` ``'response'`` ``False`` ``16``  ``'16 bit value sent'``        
    ``S32``      ``'0010'`` ``'response'`` ``False`` ``32``  ``'32 bit value sent'``           
    ``S16F``     ``'0100'`` ``'response'`` ``False`` ``16``  ``'16 bit field value sent'``           
    ``S32F``     ``'0101'`` ``'response'`` ``True``  ``32``  ``'32 bit field value sent'``                    
    ``ERROR``    ``'0111'`` ``'error'``    ``True``  ``...`` ``'Cannot run command'``                   
    ``NO_WRITE`` ``'1000'`` ``'no write'`` ``True``  ``...`` ``'No write access'``                  
    ============ ========== ============== ========= ======= =============================
    """

#   Name         Value   Mode       Indexed Bits  Description
    NONE     = ('0000', 'none',     ...,    ..., 'No response')
    S16      = ('0001', 'response', False,   16, '16 bit value sent')
    S32      = ('0010', 'response', False,   32, '32 bit value sent')
    S16F     = ('0100', 'response', True,    16, '16 bit field value sent')
    S32F     = ('0101', 'response', True,    32, '32 bit field value sent')
    ERROR    = ('0111', 'error',    ...,    ..., 'Cannot run command')
    NO_WRITE = ('1000', 'no write', ...,    ..., 'No write access')


def get_parameter_code(telegram_type, mode, indexed, bits):
    """Return the parameter code (as an enum member) that matches the 
    arguments.
    
    *telegram_type* is ``'query'`` for messages to the pump and ``'reply'`` for
    messages from the pump.
    
    Raises:
         :class:`ValueError`: If the number of matching members isn't 1,
             or if *telegram_type* is invalid.
    """
    
    if telegram_type == 'query':
        enum = ParameterAccess
    elif telegram_type == 'reply':
        enum = ParameterResponse
    else:
        raise ValueError(f'invalid telegram_type: {telegram_type}')
        
    results = []
        
    for member in enum:    
        mode_match = member.mode == mode
        index_match = member.indexed in [indexed, ...]
        bits_match = member.bits in [bits, ...]
        
        if mode_match and index_match and bits_match:
            results.append(member)
            
    if len(results) == 0:
        raise ValueError('no matching codes')
    
    if len(results) > 1:
        raise ValueError('several matching codes')
        
    return results[0]


def get_parameter_mode(telegram_type, code):
    """Return the parameter mode that matches the arguments.
    
    *telegram_type* is ``'query'`` for messages to the pump and ``'reply'`` for
    messages from the pump.
    """
    
    if telegram_type == 'query':
        enum = ParameterAccess
    elif telegram_type == 'reply':
        enum = ParameterResponse
    else:
        raise ValueError(f'invalid telegram_type: {telegram_type}')
        
    member = enum(code)
    return member.mode    


### Parameter errors ###

class ParameterExceptionMeta(type):
    """A metaclass for parameter exceptions.
        
    The instances of this metaclass are different types of parameter
    exceptions, which all have the :attr:`member` property.
    Because these instances are error classes instead of instances of error
    classes, the correct syntax to access this property is
    
    ::
        
        ParameterErrorX.member
        
    instead of 
    
    ::
        
        ParameterErrorX().member
        
    This metaclass exists, because :class:`ParameterException` subclasses and
    :class:`ParameterError` members which represent the same error need to
    contain references to each other.
    Direct references would create circular dependencies which would prevent
    the objects from being initialized.
    The easiest way to circumvent this problem is to define the
    :attr:`member` attribute as a property that is computed at runtime.
    Python doesn't have built-in class properties, so the easiest way to give
    exception classes a property is to use a metaclass. 
    """
    
    @property
    def member(self):
        """Return the :class:`ParameterException` member which has *self* as
        its :class:`description <ParameterError>` attribute.
        
        Raises:
            AttributeError: If *self* doesn't match exactly one
                :class:`ParameterError` member.
        """
        results = []
        for member in ParameterError:
            if member.exception == self:
                results.append(member)
                
        if len(results) != 1:
            raise AttributeError(f'error class matches {len(results)} members')
        
        return results[0]


class ParameterException(Exception, metaclass=ParameterExceptionMeta):
    """A superclass for exceptions that represent different error conditions 
    which are raised when the pump cannot access a parameter.
    
    This class uses the :class:`ParameterExceptionMeta` metaclass, so its
    subclasses have the :attr:`~ParameterExceptionMeta.member` attribute.
    However, this superclass itself doesn't represent any specific error, so
    trying to access :attr:`ParameterException.member
    <ParameterExceptionMeta.member>` raises an :class:`AttributeError`.
    """
    
    @property
    def member(self):
        """Return the :attr:`~ParameterExceptionMeta.member` attribute of the
        class of which *self* is an instance.
        
        This property is defined in order for the syntax 
        
        ::
            
            ParameterErrorX().member
            
        to work alongside
        
        ::
            
            ParameterErrorX.member
            
        This makes it possible to write try-catch blocks like the following: 
        
        ::
            
            try:
                # Do something.
            except ParameterException as error:
                member = error.member
                # This would have to be "type(error).member" without this
                # property.
                
                # Do something with *member*.
        """
        return type(self).member


class WrongNumError(ParameterException):
    """Raised when trying to access a parameter with a number which doesn't 
    exist.
    """


class CannotChangeError(ParameterException):
    """Raised when a parameter cannot be changed."""


class MinMaxError(ParameterException):
    """Raised when a value cannot be assigned to a parameter because it is 
    outside the range of accepted values.
    """


class ParameterIndexError(ParameterException):
    """Raised when trying to access a non-existent index of a parameter."""


class AccessError(ParameterException):
    """Raised when the access code doesn't match the parameter the pump tries 
    to access.
    """


class OtherError(ParameterException):
    """Raised for other parameter error conditions."""


class SavingError(ParameterException):
    """Raised when trying to access a parameter which is simultaneously being 
    saved to nonvolatile memory.
    """


class CustomInt(int):
    """A custom superclass for enums that inherit :class:`int` while having
    members with multiple fields.
    """
    
    def __new__(cls, value, *args, **kwargs):
        """The arguments of ``__new__`` methods in enums inheriting 
        :class:`int` are always passed to :meth:`int.__new__`, 
        which often raises an error if there are multiple arguments present.
        This class solves the problem by only passing the first argument to 
        :meth:`int.__new__`.
        """
        return int.__new__(cls, value)


class ParameterError(CustomInt, e.Enum):
    """Different parameter errors.
    
    This class also inherits :class:`CustomInt`, which means its members can
    e.g. be easily ordered. 
    
    Members of this enum have the following fields:
        **value**
            The number of the error (:class:`int`).
        **description**
            A short :class:`str` describing of the meaning of the error.
        **exception**
            The :class:`ParameterException` subclass that corresponds to this 
            member.
            
    This enum contains the following members:
    
    ================= ======= ==================================================== ============================
    Member name       Value   Description                                          Exception
    ================= ======= ==================================================== ============================
    ``WRONG_NUM``       ``0`` ``'Invalid parameter number'``                       :class:`WrongNumError`
    ``CANNOT_CHANGE``   ``1`` ``'Parameter cannot be changed'``                    :class:`CannotChangeError`
    ``MINMAX``          ``2`` ``'Min/max error'``                                  :class:`MinMaxError`
    ``INDEX``           ``3`` ``'Index error'``                                    :class:`ParameterIndexError`
    ``ACCESS``          ``5`` ``'Access mode doesn't match parameter'``            :class:`AccessError`
    ``OTHER``          ``18`` ``'Other error'``                                    :class:`OtherError`
    ``SAVING``        ``102`` ``'Parameter is being saved to nonvolatile memory'`` :class:`SavingError`
    ================= ======= ==================================================== ============================
    """

    def __init__(self, value, description, exception):
        self._value_ = value
        self.description = description
        self.exception = exception
                        
    WRONG_NUM =     (  0, 'Invalid parameter number',    WrongNumError)
    CANNOT_CHANGE = (  1, 'Parameter cannot be changed', CannotChangeError)
    MINMAX        = (  2, 'Min/max error',               MinMaxError)
    INDEX         = (  3, 'Index error',                 ParameterIndexError)
    ACCESS        = (  5, "Access mode doesn't match "
                          "parameter",                   AccessError)
    OTHER         = ( 18, 'Other error',                 OtherError)
    SAVING        = (102, 'Parameter is being saved to '
                          'nonvolatile memory',          SavingError)
    # Error codes 3, 5 and 102 aren't included in the manual, but were 
    # discovered while testing the pump.


### Control and status bits ###
    
class FlagBits(CustomInt, e.Enum):
    """A superclass for control and status bits.
    
    This class is otherwise similar to :class:`ParameterError`, but it doesn't 
    have any members since it's only meant to be subclassed, 
    and the members of its subclasses lack the **exception** field. 
    """

    def __init__(self, value, description):
        self._value_ = value
        self.description = description


class ControlBits(FlagBits):
    """Different control bits.
    
    Unused control bits have been assigned values in order to 
    prevent errors in the program.
    
    This class inherits :class:`FlagBits`, but for some reason Sphinx doesn't 
    show that properly.
    
    This enum contains the following members:
    
    =============== ====== ==================================
    Member name     Value  Description
    =============== ====== ==================================
    ``ON``           ``0`` ``'Turn or keep the pump on'``       
    ``UNUSED1``      ``1`` ``'Unknown control bit: 1'``        
    ``UNUSED2``      ``2`` ``'Unknown control bit: 2'``           
    ``UNUSED3``      ``3`` ``'Unknown control bit: 3'``                    
    ``UNUSED4``      ``4`` ``'Unknown control bit: 4'``                    
    ``X201``         ``5`` ``'Output X201 (air cooling)'``                   
    ``SETPOINT``     ``6`` ``'Enable frequency setpoint'``                  
    ``RESET_ERROR``  ``7`` ``'Reset error (all components)'``                   
    ``STANDBY``      ``8`` ``'Enable standby'``                   
    ``UNUSED9``      ``9`` ``'Unknown control bit: 9'``                   
    ``COMMAND``     ``10`` ``'Enable control bits'``                  
    ``X1_ERROR``    ``11`` ``'Error operation relay X1'``            
    ``X1_WARNING``  ``12`` ``'Normal operation relay X1'``                
    ``X1_NORMAL``   ``13`` ``'Warning relay X1'``                
    ``X202``        ``14`` ``'Output X202 (packing pump)'``                  
    ``X203``        ``15`` ``'Output X203 (venting valve)'``                                                    
    =============== ====== ==================================
    """
    # Some weird interaction of Enum, inheriting from int, and redefining
    # __new__ in FlagBits prevents Sphinx from displaying the inheritance
    # properly (it shows "Bases: <unknown>.FlagBits"), so :show-inheritance:
    # isn't used.
    ON            = ( 0, 'Turn or keep the pump on')
    """Lorem ipsum."""
    
    UNUSED1       = ( 1, 'Unknown control bit: 1')
    """Dolor sit amet."""
    
    UNUSED2       = ( 2, 'Unknown control bit: 2')
    UNUSED3       = ( 3, 'Unknown control bit: 3')
    UNUSED4       = ( 4, 'Unknown control bit: 4')
    X201          = ( 5, 'Output X201 (air cooling)') 
    SETPOINT      = ( 6, 'Enable frequency setpoint')
    RESET_ERROR   = ( 7, 'Reset error (all components)')
    STANDBY       = ( 8, 'Enable standby')
    UNUSED9       = ( 9, 'Unknown control bit: 9')
    COMMAND       = (10, 'Enable control bits')
    X1_ERROR      = (11, 'Error operation relay X1')
    X1_WARNING    = (12, 'Normal operation relay X1')
    X1_NORMAL     = (13, 'Warning relay X1')
    X202          = (14, 'Output X202 (packing pump)')
    X203          = (15, 'Output X203 (venting valve)')
    # According to the manual, bit 10 enables bits 
    # 0, 5, 6, 7, 8, 13, 14 and 15. Either bits 11 and 12 don't need 
    # to be enabled, or the manual is wrong.


class StatusBits(FlagBits):
    """Different status bits.
    
    Unused status bits have been assigned values in order to 
    prevent errors in the program.
    
    This class inherits :class:`FlagBits`, but for some reason Sphinx doesn't 
    show that properly.
    
    This enum contains the following members:
    
    =================== ====== ======================================
    Member name         Value  Description
    =================== ====== ======================================
    ``READY``            ``0`` ``'Ready for operation'``       
    ``UNUSED1``          ``1`` ``'Unknown status bit: 1'``        
    ``OPERATION``        ``2`` ``'Operation enabled'``           
    ``ERROR``            ``3`` ``'Error condition (all components)'``           
    ``ACCELERATION``     ``4`` ``'Accelerating'``                    
    ``DECELERATION``     ``5`` ``'Decelerating'``                   
    ``SWITCH_ON_LOCK``   ``6`` ``'Switch-on lock'``                  
    ``TEMP_WARNING``     ``7`` ``'Temperature warning'``                   
    ``UNUSED8``          ``8`` ``'Unknown status bit: 8'``                   
    ``PARAM_CHANNEL``    ``9`` ``'Parameter channel enabled'``                   
    ``DETAINED``        ``10`` ``'Normal operation detained'``                  
    ``TURNING``         ``11`` ``'Pump is turning'``            
    ``UNUSED12``        ``12`` ``'Unknown status bit: 12'``                
    ``OVERLOAD``        ``13`` ``'Overload warning'``                
    ``WARNING``         ``14`` ``'Collective warning'``                  
    ``PROCESS_CHANNEL`` ``15`` ``'Process channel enabled'``                                                    
    =================== ====== ======================================
    """    
    READY           = ( 0, 'Ready for operation')
    UNUSED1         = ( 1, 'Unknown status bit: 1')
    OPERATION       = ( 2, 'Operation enabled')
    ERROR           = ( 3, 'Error condition (all components)')
    ACCELERATION    = ( 4, 'Accelerating')
    DECELERATION    = ( 5, 'Decelerating')
    SWITCH_ON_LOCK  = ( 6, 'Switch-on lock')
    TEMP_WARNING    = ( 7, 'Temperature warning')
    UNUSED8         = ( 8, 'Unknown status bit: 8')
    PARAM_CHANNEL   = ( 9, 'Parameter channel enabled')
    DETAINED        = (10, 'Normal operation detained')
    TURNING         = (11, 'Pump is turning')
    UNUSED12        = (12, 'Unknown status bit: 12')
    OVERLOAD        = (13, 'Overload warning')
    WARNING         = (14, 'Collective warning')
    PROCESS_CHANNEL = (15, 'Process channel enabled')


"""Unit tests for the output module."""

import unittest

from turboctl import (help_string, full_output, parameter_output, 
                      control_or_status_output, hardware_output, Query, Reply, 
                      ParameterError, ControlBits, StatusBits, Command)

class Base(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        # Print long strings.
        cls.maxDiff = None

class TestOther(Base):
    
    def test_full_output(self):
        
        q = Query(parameter_number=1,
                  parameter_value=123,
                  parameter_mode='write',
                  control_or_status_set={ControlBits.COMMAND,
                                         ControlBits.ON,
                                         ControlBits.FREQ_SETPOINT},
                   frequency=1000
                   )
        
        r = Reply(parameter_number=1,
                  parameter_value=123,
                  parameter_mode='response',
                  control_or_status_set={StatusBits.TURNING,
                                         StatusBits.ACCELERATION},
                   frequency=1000,
                   temperature=30,
                   current=100,
                   voltage=120
                   )

        string = (
"""
Sent a telegram with the following contents:
    Write the value 123 to parameter 1
    Active control bits:
        Start/Stop
        Set frequency setpoint
        Enable control bits 0, 5, 6, 7, 8, 13, 14, 15
    Stator frequency setpoint: 1000 Hz

Received a telegram with the following contents:
    The value of parameter 1 is 123
    Present status conditions:
        Accelerating
        Pump is turning
    Stator frequency: 1000 Hz
    Frequency converter temperature: 30 °C
    Motor current: 100×0.1 A
    Intermediate circuit voltage: 120×0.1 V
""".strip()                
)
        self.assertEqual(full_output(q, r), string)
    
    def test_help_string(self):
        commands = [
            Command(names=['test1', 't, t1'],
                function='cmd_test1',
                args=[],
                description='Test command 1'),
        
            Command(names=['test2', 't2'],
                    function='cmd_test2',
                    args=['a', 'b', 'c=1'],
                    description=('Test command 2 with a very long '
                                 'description. This long description is '
                                 'intended to test line wrapping.'))
        ]
        
        string = (
"""
Accepted commands:
Command    Aliases    Args           Description
test1      t, t1                     Test command 1
test2      t2         <a> <b> [c=1]  Test command 2 with a very
                                     long description. This long
                                     description is intended to
                                     test line wrapping.
""".strip()
)
        self.assertEqual(help_string(commands), string)

class TestParameterOutput(Base):
    
    def setUp(self):
        self.q    = Query(parameter_number=1,  parameter_value=123)
        self.q_Hz = Query(parameter_number=3,  parameter_value=123)
        self.qi   = Query(parameter_number=29, parameter_value=123, 
                          parameter_index=1)
        self.r    = Reply(parameter_number=1,  parameter_value=123)
        self.r_Hz = Reply(parameter_number=3,  parameter_value=123)
        self.ri   = Reply(parameter_number=29, parameter_value=123, 
                          parameter_index=1)
    
    def test_read(self):
        self.q.parameter_mode = 'read'
        self.qi.parameter_mode = 'read'

        string = 'Return the value of parameter 1'
        indexed_string = 'Return the value of parameter 29, index 1'
        
        self.assertEqual(parameter_output(self.q), string)
        self.assertEqual(parameter_output(self.qi), indexed_string)
        self.assertEqual(parameter_output(self.q, verbose=False), '')    
        
    def test_write(self):
        self.q.parameter_mode = 'write'
        self.qi.parameter_mode = 'write'
        self.q_Hz.parameter_mode = 'write'
        
        string = 'Write the value 123 to parameter 1'
        index_string = 'Write the value 123 to parameter 29, index 1'
        unit_string = 'Write the value 123 Hz to parameter 3'
        
        self.assertEqual(parameter_output(self.q), string)
        self.assertEqual(parameter_output(self.qi), index_string)
        self.assertEqual(parameter_output(self.q_Hz), unit_string)
        self.assertEqual(parameter_output(self.q, verbose=False), '')
        
    def test_none(self):
        q = Query()
        r = Reply()
        
        string = 'No parameter access'
        self.assertEqual(parameter_output(q), string)
        self.assertEqual(parameter_output(r), string)
        self.assertEqual(parameter_output(q, verbose=False), '')
        self.assertEqual(parameter_output(r, verbose=False), '')
        
    def test_response(self):
        self.r.parameter_mode = 'response'
        self.ri.parameter_mode = 'response'
        self.r_Hz.parameter_mode = 'response'
        
        string = 'The value of parameter 1 is 123'
        index_string = 'The value of parameter 29, index 1 is 123'
        unit_string = 'The value of parameter 3 is 123 Hz'
        
        self.assertEqual(parameter_output(self.r), string)
        self.assertEqual(parameter_output(self.ri), index_string)
        self.assertEqual(parameter_output(self.r_Hz), unit_string)
        self.assertEqual(parameter_output(self.r, verbose=False), '123')
        self.assertEqual(parameter_output(self.r_Hz, verbose=False), '123 Hz')
        
    def test_error(self):
        self.r.parameter_mode = 'error'
        self.ri.parameter_mode = 'error'
        self.r.error_code = ParameterError.MINMAX.value
        self.ri.error_code = ParameterError.MINMAX.value
        
        string = "Can't access parameter 1: min/max error"
        index_string = (
            "Can't access parameter 29, index 1: min/max error")
        
        self.assertEqual(parameter_output(self.r), string)
        self.assertEqual(parameter_output(self.ri), index_string)
        self.assertEqual(parameter_output(self.r, verbose=False), 
                         'Error: min/max restriction')
        
    def test_no_write(self):
        self.r.parameter_mode = 'no write'
        self.ri.parameter_mode = 'no write'
        
        string = "Parameter 1 isn't writable"
        index_string = "Parameter 29 isn't writable"
        
        self.assertEqual(parameter_output(self.r), string)
        self.assertEqual(parameter_output(self.ri), index_string)
        self.assertEqual(parameter_output(self.r, verbose=False), 
                         'Not writable')


# Shorten lines by defining an alias:
out = control_or_status_output

class TestControlOrStatusOutput(Base):
    
    def test_normal(self):
        
        self.q = Query(control_or_status_set = {ControlBits.ON, 
                                                   ControlBits.COMMAND})
        self.r = Reply(control_or_status_set = {StatusBits.TURNING, 
                                                StatusBits.ACCELERATION})

        control_string = ('Active control bits:\n'
                          '    Start/Stop\n'
                          '    Enable control bits')
        status_string = ('Present status conditions:\n'
                         '    Accelerating\n'
                         '    Pump is turning')
        
        self.assertEqual(out(self.q), control_string)
        self.assertEqual(out(self.r), status_string)
        self.assertEqual(out(self.q, verbose=False), '')    
        self.assertEqual(out(self.r, verbose=False), '')    
        
    def test_errors_and_warnings(self):
        """Make sure control_or_status_output(wrapper, verbose=False) 
        doesn't return '' if there are errors or warnings present.
        """
        
        self.error   = Reply(control_or_status_set = {StatusBits.ERROR})
        self.warning = Reply(control_or_status_set = {StatusBits.WARNING})
        self.both    = Reply(control_or_status_set = {StatusBits.ERROR, 
                                                      StatusBits.WARNING})
        
        error_string = 'Error(s) present'
        warning_string = 'Warning(s) present'
        both_string = 'Error(s) present\nWarning(s) present'
        
        self.assertEqual(out(self.error, verbose=False), error_string)    
        self.assertEqual(out(self.warning, verbose=False), warning_string)    
        self.assertEqual(out(self.both, verbose=False), both_string)
        
        
class TestHardwareOutput(unittest.TestCase):
    
    def test_normal(self):
        q = Query(frequency=1000, temperature=30, current=100, voltage=120)
        r = Reply(frequency=1000, temperature=30, current=100, voltage=120)
        
        long_string = (
            'Stator frequency: 1000 Hz\n'
            'Frequency converter temperature: 30 °C\n'
            'Motor current: 100×0.1 A\n'
            'Intermediate circuit voltage: 120×0.1 V')
        
        short_string = (
            'f=1000 Hz\n'
            'T=30 °C\n'
            'I=100×0.1 A\n'
            'U=120×0.1 V')
        
        self.assertEqual(hardware_output(r), long_string)    
        self.assertEqual(hardware_output(r, verbose=False), short_string)   
        self.assertEqual(hardware_output(q), '')    
        self.assertEqual(hardware_output(q, verbose=False), '')    
        
    def test_setpoint(self):
        q = Query(frequency=1000, temperature=30, current=100, voltage=120, 
                  control_or_status_set = {ControlBits.COMMAND, 
                                           ControlBits.SETPOINT})
        
        string = ('Stator frequency setpoint: 1000 Hz')
        
        self.assertEqual(hardware_output(q), string)    
        self.assertEqual(hardware_output(q, verbose=False), '')    
    
    
if __name__ == '__main__':
    unittest.main()


"""This module contains some temporary inelegant fixes that hopefully can be
removed later.
"""
import weakref

import urwid


def monkeypatch_urwid():
    """Monkey-patch a bug in Urwid.

    Urwid has a bug that causes a memory leak; the issue is documented here:
    https://github.com/urwid/urwid/issues/451
    The issue is still open on 2024-05-07 and apparently the simple fix here
    breaks something else, but it seems to work for the purposes of TurboCtl. 
    """

    # The CanvasCache.store method, copied from urwid, with only one line
    # altered.
    # It might be possible to only patch that single line by modifying the code
    # with the ast or inspect modules, but this is much easier. 
    def store(cls, wcls, canvas):
        """
        Store a weakref to canvas in the cache.
    
        wcls -- widget class that contains render() function
        canvas -- rendered canvas with widget_info (widget, size, focus)
        """
        if not canvas.cacheable:
            return
    
        assert canvas.widget_info, "Can't store canvas without widget_info"
        widget, size, focus = canvas.widget_info
        def walk_depends(canv):
            """
            Collect all child widgets for determining who we
            depend on.
            """
            # FIXME: is this recursion necessary?  The cache
            # invalidating might work with only one level.
            depends = []
            for x, y, c, pos in canv.children:
                if c.widget_info:
                    depends.append(c.widget_info[0])
                elif hasattr(c, 'children'):
                    depends.extend(walk_depends(c))
            return depends
    
        # use explicit depends_on if available from the canvas
        depends_on = getattr(canvas, 'depends_on', None)
        if depends_on is None and hasattr(canvas, 'children'):
            depends_on = walk_depends(canvas)
        if depends_on:
            for w in depends_on:
                if w not in cls._widgets:
                    return
            for w in depends_on:
                # The patched part is here.
                # The original line:
                # cls._deps.setdefault(w,[]).append(widget)
                # The patched line: 
                cls._deps.setdefault(w,set()).add(widget)
    
        ref = weakref.ref(canvas, cls.cleanup)
        cls._refs[ref] = (widget, wcls, size, focus)
        cls._widgets.setdefault(widget, {})[(wcls, size, focus)] = ref
    store = classmethod(store)
    
    # Replace the original store method with the modified version.
    urwid.CanvasCache.store = store


"""This module controls how the status of the pump is presented to the user in
the UI.

Attributes:
    palette:
        A colour palette for an
        :class:`~turboctl.ui.advanced_tui.AdvancedTUI`.

    red_button:
        This is displayed by :class:`urwid.Text` widgets as a red
        circle, if the widgets are used in an
        :class:`~turboctl.ui.advanced_tui.AdvancedTUI`
        initialized with :attr:`palette`.

    green_button:
        Similar to :attr:`red_button`, but green instead of red.
"""

# (name, text color, background color)
palette = [('green', 'light green', ''),
           ('red', 'light red', '')]

red_button = ('red', '⏺')
green_button = ('green', '⏺')
grey_button = ('grey', '⏺')


def status_screen(status):
    """Return a status screen for the
    :attr:`display <turboctl.ui.advanced_tui.AdvancedTUI.display>`
    of an :class:`~turboctl.ui.advanced_tui.AdvancedTUI`.

    Args:
        status:
            A :class:`~turboctl.ui.control_interface.api.Status` object, the
            contents of which will be displayed on the screen.

    Returns:
        A nested iterable of strings, which will be interpreted by
        :class:`urwid.Text` as coloured text.
        The format is explained
        `here <http://urwid.org/manual/displayattributes.html
        ?highlight=display%20attributes#text-markup>`_.
    """
    match status.pump_on:
        case None:
            onoff_str = 'Pump status unknown'
            button = grey_button
        case True:
            onoff_str = 'Pump on'
            button = green_button
        case False:
            onoff_str = 'Pump off'
            button = red_button
        case _:
            raise ValueError(
                f'status.pump_on should be None, True or False, '
                f'was {status.pump_on}')

    if status.status_bits is None:
        lines = None
    else:
        descriptions = (member.description for member in status.status_bits)        
        lines = ('    ' + s for s in descriptions)

    match lines:
        case None:
            condition_str = 'Status conditions unknown'
        case []:
            condition_str = 'No active status conditions'
        case _:
            condition_str = 'Active status conditions:\n' + '\n'.join(lines)        

    def format_float(x, unit, format_str='{}'):
        if x is None:
            return 'unknown'
        return f'{format_str.format(x)} {unit}'

    hardware_str = (
        'Frequency: ' + format_float(status.frequency, 'Hz') + '\n'
        'Temperature: ' + format_float(status.temperature, '°C') + '\n'
        'Current: ' + format_float(status.current, 'A', '{:.1f}') + '\n'
        'Voltage: ' + format_float(status.voltage, 'V')
    )

    text = ([
        onoff_str, ' ', button, '\n',
        condition_str, '\n',
        hardware_str
    ])
    return text


"""Unit tests for the telegram module."""

import unittest

from turboctl.telegram.codes import (ParameterResponse, ParameterError,
                                     ControlBits)
from turboctl.telegram.datatypes import Uint, Sint, Float, Bin
from turboctl.telegram.telegram import (Telegram, TelegramBuilder, 
                                        TelegramReader, checksum)
from test_turboctl.telegram.test_parser import dummy_parameter


# The actual parameters are replaced with dummy values for easier testing.
# TelegramBuilder is called with the parameters argument everywhere in this
# module for simplicity, even though not all tests care about which parameters
# are used.
parameters = {
    # A basic parameter.
    1234: dummy_parameter(number=1234),
    # Parameters of different types.
    1: dummy_parameter(number=1, datatype=Sint),
    2: dummy_parameter(number=2, datatype=Float),
    3: dummy_parameter(number=3, datatype=Bin),
    # An indexed parameter with a different number of bits. 
    4: dummy_parameter(number=4, indices=range(1), bits=32),
}


class Base(unittest.TestCase):
    
    def setUp(self):
        self.kwargs = {
            'parameter_code': Bin('0001'),
            'parameter_number': Uint(1234, 11), # '0001' + '0' + 1234 = 20 210
            'parameter_index': Uint(56),
            'parameter_value': Uint(5432, 32), # = 21 56
            'flag_bits': Bin('1111000011001100'), # = 51 15
            'frequency': Uint(1221, 16), # = 4 197
            'temperature': Sint(2332, 16), # = 9 28
            'current': Uint(3443, 16), # = 13 115
            'voltage': Uint(4554, 16) # = 17 202
        }
        
        self.telegram = Telegram(**self.kwargs)
        
        self.bytes = bytes([
            2, 22, 0,     # Start 
            20, 210,      # Parameter code and number
            0,            # Empty
            56,           # Parameter index
            0, 0, 21, 56, # Value
            51, 15,       # Flag bits
            4, 197,       # Frequency
            9, 28,        # Temperature
            13, 115,      # Current
            0, 0,         # Empty
            17, 202,      # Voltage
            138           # Checksum
        ])


class TestTelegram(Base):
    
    def test_bytes(self):
        self.assertEqual(bytes(self.telegram), self.bytes)
        
    def _test_bytes_with_value(self, value):
        # test_bytes, but telegram.parameter_value is replaced with *value*.
        
        # Replace parameter_value in self.telegram.
        self.kwargs['parameter_value'] = value
        
        # Replace parameter_value in self.bytes.
        self.bytes = self.bytes[:7] + bytes(value) + self.bytes[11:]
        
        # Update the checksum.
        self.bytes = self.bytes[:-1] + bytes([checksum(self.bytes[:-1])])
        
        t = Telegram(**self.kwargs)
        self.assertEqual(bytes(t), self.bytes)

    def test_bytes_sint(self):
        value = Sint(-1234, 32)        
        self._test_bytes_with_value(value)
        
    def test_bytes_float(self):
        value = Float(1234.5678)
        self._test_bytes_with_value(value)
        
    def test_bytes_bin(self):        
        value = Bin(4 * '10010110')
        self._test_bytes_with_value(value)


class TestTelegramBuilder(Base):
                
    ### Setters ###
            
    def test_setters(self):
        
        # Test creating a query.
        
        tb = (TelegramBuilder(parameters)
            # 0001 is "read a value".
            .set_parameter_mode('read')
            .set_parameter_number(1234)
            .set_parameter_index(56)
            .set_parameter_value(5432)
            .set_flag_bits(
            [ControlBits(i) for i, c 
             in enumerate('1111000011001100') if c == '1'])
            .set_frequency(1221)
            .set_temperature(2332)
            .set_current(3443)
            .set_voltage(4554))
        
        self.assertEqual(tb.build('query'), self.telegram)

        # Test creating a reply.

        # 0001 is "16 bit value sent".
        tb.set_parameter_mode('response')
        self.assertEqual(tb.build('reply'), self.telegram)
        
    def test_default(self):
        t = Telegram(
            parameter_code=Bin('0000'), 
            parameter_number=Uint(0, 11), 
            parameter_index=Uint(0, 8), 
            parameter_value=Uint(0, 32), 
            flag_bits=Bin(16 * '0'), 
            frequency=Uint(0, 16), 
            temperature=Sint(0, 16), 
            current=Uint(0, 16),
            voltage=Uint(0, 16)
        )
        self.assertEqual(TelegramBuilder(parameters).build(), t)
        
    def test_wrong_type_raises_value_error(self):
        with self.assertRaises(ValueError):
            TelegramBuilder(parameters).build('qery')
            
    ### from_bytes ###
    
    def test_from_bytes(self):
        t = TelegramBuilder(parameters).from_bytes(self.bytes).build()
        self.assertEqual(t, self.telegram)
            
    def test_from_bytes_with_invalid_byte_0_fails(self):
        bytes2 = bytes([0]) + self.bytes[1:]
        with self.assertRaises(ValueError):
            TelegramBuilder(parameters).from_bytes(bytes2)
    
    def test_from_bytes_with_invalid_byte_1_fails(self):
        bytes2 = self.bytes[0:1] + bytes([0]) + self.bytes[2:]
        with self.assertRaises(ValueError):
            TelegramBuilder(parameters).from_bytes(bytes2)
            
    def test_from_bytes_with_invalid_length_fails(self):
        bytes2 = self.bytes + bytes([0]) 
        with self.assertRaises(ValueError):
            TelegramBuilder(parameters).from_bytes(bytes2)
            
    def test_from_bytes_with_invalid_checksum_fails(self):
        bytes2 = self.bytes[:-1] + bytes([0])
        with self.assertRaises(ValueError):
            TelegramBuilder(parameters).from_bytes(bytes2)
            
    ### Parameter value ###
    
    def test_parameter_value_sint(self):
        t = (TelegramBuilder(parameters)
            .set_parameter_number(1)
            .set_parameter_mode('write')
            .set_parameter_value(-1234)
            .build())
        
        self.assertEqual(t.parameter_value, Sint(-1234, 32))
        
    def test_parameter_value_float(self):
        t = (TelegramBuilder(parameters)
            .set_parameter_number(2)
            .set_parameter_mode('write')
            .set_parameter_value(1234.5678)
            .build())

        self.assertEqual(t.parameter_value, Float(1234.5678))
        
    def test_parameter_value_int_as_float(self):
        t = (TelegramBuilder(parameters)
            .set_parameter_number(2)
            .set_parameter_mode('write')
            .set_parameter_value(1234)
            .build())

        self.assertEqual(t.parameter_value, Float(1234.0))
        
    def test_parameter_value_bin(self):
        t = (TelegramBuilder(parameters)
            .set_parameter_number(3)
            .set_parameter_mode('write')
            .set_parameter_value(4 * '10010110')
            .build())

        self.assertEqual(t.parameter_value, Bin(4 * '10010110'))
        
    def test_parameter_value_int_as_bin(self):
        t = (TelegramBuilder(parameters)
            .set_parameter_number(3)
            .set_parameter_mode('write')
            .set_parameter_value(2526451350)
            .build())

        self.assertEqual(t.parameter_value, Bin(4 * '10010110'))        
        
    def test_parameter_value_invalid_value_fails(self):
        tb = (TelegramBuilder(parameters)
            .set_parameter_number(1234)
            .set_parameter_mode('write')
            .set_parameter_value(-1234)
        )
        with self.assertRaises(ValueError):
            tb.build()
        
    def test_parameter_value_wrong_type_fails(self):
        tb = (TelegramBuilder(parameters)
            .set_parameter_number(3)
            .set_parameter_mode('write')
            .set_parameter_value(12.34)
        )
        with self.assertRaises(TypeError):
            tb.build()
            
    def test_parameter_value_with_no_parameter_access(self):
        """Make sure the parameter type is always Uint when there is no
        parameter access.
        """
        for mode in ['none', 'error']:
            with self.subTest(i=mode):
        
                t =  (TelegramBuilder(parameters)
                     .set_parameter_number(3)
                     .set_parameter_mode('none')
                     .set_parameter_value(1234)
                     .build())
                self.assertEqual(t.parameter_value, Uint(1234, 32))
            
    ### Parameter value from bytes
    
    # The most basic case has already been tested in test_from_bytes.
            
    def _test_value_from_bytes(self, param_number, param_value):
        # test_from_bytes, but telegram.parameter_number and 
        # telegram.parameter_value are replaced with *param_number* and
        # *param_value*.
        
        parameter_bytes = bytes(Bin('00010') + Uint(param_number, 11))
        value_bytes = bytes(param_value)
        
        # Replace parameter_number and parameter_value in self.bytes.
        self.bytes = (self.bytes[:3] + parameter_bytes + self.bytes[5:7]
                      + value_bytes + self.bytes[11:])
        
        # Update the checksum.
        self.bytes = self.bytes[:-1] + bytes([checksum(self.bytes[:-1])])
        
        t = TelegramBuilder(parameters).from_bytes(self.bytes).build()
        self.assertEqual(t.parameter_value, param_value)
        
    def test_sint_from_bytes(self):
        number = 1
        value = Sint(-1234, 32)        
        self._test_value_from_bytes(number, value)
        
    def test_float_from_bytes(self):        
        number = 2
        value = Float(1234.5678)
        self._test_value_from_bytes(number, value)        
        
    def test_bin_from_bytes(self):        
        number = 3
        value = Bin(4 * '10010110')
        self._test_value_from_bytes(number, value)        
            
    ### Parameter code ###
    
    def test_parameter_codes(self):
        # Test some parameter codes with an indexed 32 bit parameter.
        # An unindexed 16 bit prameter was tested in test_setters.
        
        tb = (TelegramBuilder(parameters).set_parameter_number(4))
        
        tb.set_parameter_mode('read')
        t = tb.build('query')
        self.assertEqual(t.parameter_code, Bin('0110'))

        tb.set_parameter_mode('write')
        t = tb.build('query')
        self.assertEqual(t.parameter_code, Bin('1000'))


        tb.set_parameter_mode('response')
        t = tb.build('reply')
        self.assertEqual(t.parameter_code, Bin('0101'))
        
    ### Parameter code from bytes ###
        
    def _set_parameter_access_and_value(self, code, number, value):
        """Update the parameter code, number and value in self.bytes.
        
        The checksum is updated automatically.
        
        Args:
            code: A string.
            number: An int.
            value: A Data subclass instance.
        """
        # Change the 4 code bits but keep the other bits in the parameter
        # block the same.
        parameter_block = bytes(Bin(code) + Bin('0') + Uint(number, 11))

        # Replace the parameter code/number and value blocks.
        self.bytes = (self.bytes[:3] + parameter_block + self.bytes[5:7]
                      + bytes(value) + self.bytes[11:])
        
        # Update the checksum.
        self.bytes = (
            self.bytes[:-1] + bytes([checksum(self.bytes[:-1])]))
    
    def test_from_bytes_no_parameter_access(self):
        """Make sure invalid parameter numbers are accepted if the parameter
        mode is 'none' or 'error'.
        """
        no_access_code = ParameterResponse.NONE.value
        error_code = ParameterResponse.ERROR.value
        
        for code in [no_access_code, error_code]:
            with self.subTest(i=code):
                
                # parameter_number is 0 (an invalid value), but this should be
                # ok since the parameter isn't accessed.
                self._set_parameter_access_and_value(code, 0, Sint(-1, 32))
                                
                # The 'error' mode only works here if the telegram is a reply.
                t = (TelegramBuilder(parameters)
                     .from_bytes(self.bytes)
                     .build('reply'))

                # The value will be interpreted as an Uint instead of a Sint,
                # since the type of the parameter wasn't checked.
                self.assertEqual(t.parameter_value, Uint(2**32 - 1, 32))
        
    def test_from_bytes_invalid_parameter_number(self):
        """Make sure invalid parameter numbers are not accepted if the
        parameter mode is not 'none' or 'error'.
        """
        # parameter_number is 0, which should raise an error, since the
        # parameter is now accessed.        
        self._set_parameter_access_and_value('0001', 0, Uint(0, 32))
        
        with self.assertRaises(ValueError):
            TelegramBuilder(parameters).from_bytes(self.bytes).build()


class TestTelegramReader(Base):
        
    ### Getters ###
    
    def test_values(self):
        tr = TelegramReader(self.telegram)
        
        self.assertEqual(tr.parameter_mode, 'response')
        self.assertEqual(tr.parameter_number, 1234)
        self.assertEqual(tr.parameter_index, 56)
        self.assertEqual(tr.parameter_value, 5432)
        self.assertEqual(
            tr.flag_bits, [ControlBits(i) for i, c 
                           in enumerate('1111000011001100') if c == '1'])
        self.assertEqual(tr.frequency, 1221)
        self.assertEqual(tr.temperature, 2332)
        self.assertEqual(tr.current, 3443)
        self.assertEqual(tr.voltage, 4554)
        
    def test_query(self):
        tr = TelegramReader(self.telegram, 'query')        
        self.assertEqual(tr.parameter_mode, 'read')
        
    def test_invalid_type_fails(self):
        with self.assertRaises(ValueError):
            TelegramReader(self.telegram, 'qery')
            
    ### parameter_mode ###
    
    # Test the same modes as in TestTelegramBuilder.test_parameter_codes.
    
    def test_parameter_mode_read(self):        
        self.kwargs['parameter_code'] = Bin('0110')
        
        t = Telegram(**self.kwargs)
        tr = TelegramReader(t, 'query')
        
        self.assertEqual(tr.parameter_mode, 'read')
        
    def test_parameter_mode_write(self):
        self.kwargs['parameter_code'] = Bin('1000')
        
        t = Telegram(**self.kwargs)
        tr = TelegramReader(t, 'query')
        
        self.assertEqual(tr.parameter_mode, 'write')
        
    def test_parameter_mode_response(self):
        self.kwargs['parameter_code'] = Bin('0101')
        
        t = Telegram(**self.kwargs)
        tr = TelegramReader(t, 'reply')
        
        self.assertEqual(tr.parameter_mode, 'response')

    ### parameter_error ###
        
    def test_parameter_error(self):
        t = (TelegramBuilder(parameters)
            .set_parameter_number(1234)
            .set_parameter_mode('error')
            .set_parameter_value(0)
            .build('reply'))
            
        tr = TelegramReader(t)
        self.assertEqual(tr.parameter_error, ParameterError.WRONG_NUM)

    def test_no_parameter_error(self):
        t = (TelegramBuilder(parameters).build('reply'))
            
        tr = TelegramReader(t)
        self.assertEqual(tr.parameter_error, None)
        
    def test_invalid_parameter_error_number_fails(self):
        t = (TelegramBuilder(parameters)
            .set_parameter_number(1234)
            .set_parameter_mode('error')
            .set_parameter_value(999)
            .build('reply'))
            
        tr = TelegramReader(t)
        with self.assertRaises(ValueError):
            tr.parameter_error
            
    ### Magic methods ###
    
    def test_repr(self):
        # Replace the default status bits since there are too many.
        self.telegram.flag_bits = Bin('1010000000000000') 
        
        tr = TelegramReader(self.telegram)
        string = f"TelegramReader(telegram={str(self.telegram)}, type='reply')"
        
        self.assertEqual(repr(tr), string)
    
    def test_str(self):
        # Replace the default status bits since there are too many.
        self.telegram.flag_bits = Bin('1010000000000000') 
        # Set the response code to 'error' so that we can make sure
        # parameter_error is printed correctly (it initially was not, because
        # f'{parameter_error}' produces a number instead of
        # 'ParameterError.X'). 
        self.telegram.parameter_code = Bin('0111')
        # Set parameter_value to a valid error code (= min/max error).
        self.telegram.parameter_value = Uint(2, 32)

        tr = TelegramReader(self.telegram)
        
        string = f"""
TelegramReader(
    telegram={str(self.telegram)},
    type='reply',
    parameter_mode='error',
    parameter_number=1234,
    parameter_index=56,
    parameter_value=2,
    parameter_error=ParameterError.MINMAX,
    flag_bits=[<StatusBits.READY: 0>, <StatusBits.OPERATION: 2>],
    frequency=1221,
    temperature=2332,
    current=3443,
    voltage=4554
)
"""[1:-1]

        self.assertEqual(str(tr), string)


if __name__ == '__main__':
    unittest.main()



"""This module was created to test the run_tests module.

When the this module is tested with run_test.run_tests, the results should 
show it as having the following test results:
    Successes:  1
    Failures:   2
    Errors:     3
    Skipped:    4
    Total:     10
"""
import unittest

class TestClass(unittest.TestCase):
        
    def test_this_should_succeed(self):
        self.assertTrue(True)
        
    def test_this_should_fail_1_of_2(self):
        self.assertTrue(False)
        
    def test_this_should_fail_2_of_2(self):
        self.assertTrue(False)

    def test_this_should_raise_an_error_1_of_3(self):
        raise ValueError('This is a test error.')
    
    @unittest.skip('#1 of 4 skipped tests')
    def test_this_should_be_skipped_1_of_4(self):
        self.assertTrue(False)
        
    @unittest.skip('#2 of 4 skipped tests')
    def test_this_should_be_skipped_2_of_4(self):
        self.assertTrue(False)        
        
        
class AnotherTestClass(unittest.TestCase):
    
    def test_this_should_raise_an_error_2_of_3(self):
        raise ValueError('This is a test error.')
        
    def test_this_should_raise_an_error_3_of_3(self):
        raise ValueError('This is a test error.')
    
    
@unittest.skip('This entire class should be skipped')
class SkippedClassWithTwoTests(unittest.TestCase):
    
    def test_this_should_be_skipped_3_of_4(self):
        self.assertTrue(False)
        
    def test_this_should_be_skipped_4_of_4(self):
        self.assertTrue(False)
        
        
if __name__ == '__main__':
    unittest.main()



"""This module contains a parser that reads data about pump parameters, 
errors and warnings from text files.

Attributes:
    PARAMETERS:
        The pump has multiple parameters which affect its behaviour.
        This attribute holds a :class:`dict` of these parameters, represented 
        as :class:`Parameter` objects, with the numbers of the parameters as 
        the keys.
        
    ERRORS:
        A :class:`dict` of different error conditions which can affect the 
        pump, represented as :class:`ErrorOrWarning` objects, 
        with error numbers as the keys. 
        
    WARNINGS:
        Like :attr:`ERRORS`, but for warnings instead of errors.
"""

import re
import ast
import os.path
from dataclasses import dataclass
from collections import OrderedDict
from typing import Union, List

from turboctl.telegram.datatypes import Data, Uint, Sint, Float


@dataclass
class Parameter:
    """A class for representing pump parameters.
    
    This is a :obj:`~dataclasses.dataclass`, so the ``__init__``, ``__str__``
    and ``__repr__`` methods are generated automatically.
    """
    
    number: int
    """The number of the parameter."""
    
    name: str
    """The name of the parameter."""
    
    indices: range
    """A :class:`range` object describing the indices of the parameter; e.g.
    ``range(5)`` would mean the parameter has indices numbered 0 to 4.
    This is ``range(0)`` for unindexed parameters.
    """

    min_value: Union[Data, str]
    """The minimum value of the parameter.
    
    The minumum value of some parameters depends on the current value of
    another parameter. In that case, :attr:`min_value` is set to
    ``'P<number>'``; e.g. 'P18' means the value of parameter 18.
    """
    
    max_value: Union[Data, str]
    """The maximum value of the parameter.
    
    The maximum value of some parameters depends on the current value of
    another parameter. In that case, :attr:`max_value` is set to
    ``'P<number>'``; e.g. 'P18' means the value of parameter 18.
    """
    
    default: Union[Data, List[Data]]
    """The default value of the parameter.
    
    If the parameter is indexed and different indices have different default
    values, :attr:`default` will be a :class:`list` of default values
    corresponding to the different indices.
    """

    unit: str
    """The unit of the parameter; e.g. ``'°C'`` for degrees Celsius."""
    
    writable: bool
    """Signifies whether the value of the parameter can be changed."""
    
    datatype: type
    """The type of the value of the parameter; a
    :class:`~turboctl.telegram.datatypes.Data` subclass
    (but not :class:`~turboctl.telegram.datatypes.Bin`).
    """
    
    bits: int
    """The size of the parameter in bits; ``16`` or ``32``."""
    
    description: str
    """A string describing the parameter."""
    
    @property
    def fields(self):
        """Return an :class:`~collections.OrderedDict` with the attribute
        names of this objects as keys and their values as values.
        """
        fieldnames = ['number', 'name', 'indices', 'min_value', 'max_value', 
                      'default', 'unit', 'writable', 'datatype', 'bits', 
                      'description']
        return OrderedDict((name, getattr(self, name)) for name in fieldnames)
    
    
@dataclass
class ErrorOrWarning:
    """A class for representing pump errors and warnings."""
    
    number: int
    """The number of the error or warning."""
    
    name: str
    """The name of the error or warning."""
    
    possible_cause: str
    """A string describing the possible cause(s) of the error or warning."""
    
    remedy: str
    """A string describing possible remedies for the error or warning."""
    
    @property
    def fields(self):
        """Return an :class:`~collections.OrderedDict` with the attribute
        names of this object as keys and their values as values.
        """
        fieldnames = ['number', 'name', 'possible_cause', 'remedy']
        return OrderedDict((name, getattr(self, name)) for name in fieldnames)


# The special character used for comments:
_COMMENT_CHAR = '#'


def main():
    """Define :attr:`PARAMETERS`, :attr:`ERRORS` and :attr:`WARNINGS`.
    
    This function is automatically executed when this module is 
    imported or run as a script.
    """
    global PARAMETERS, ERRORS, WARNINGS
    PARAMETERS = load_parameters()
    ERRORS = load_errors()
    WARNINGS = load_warnings()

def _fullpath(filename):
    """Return of the full path of *filename*. 
    *filename* should be a file in the same directory as this module.
    """
    dirpath = os.path.dirname(__file__)
    return os.path.join(dirpath, filename)


def load_parameters(path=None):
    """Parse the parameter file and return a dictionary of 
    parameters.
    
    This function is automatically called by :func:`main`, but it can also be
    called separately in order to create another set of parameters for testing
    purposes.  
    
    Args:
        path: A text file containing parameter/error/warning data.
            If *path* isn't specified, the file will be ``'parameters.txt'``,
            located in the same directory as this module.
            The syntax used to define the parameters is explained in
            ``parameters.txt``.
            
    Returns:
        A :class:`dict` with numbers as keys and :class:`Parameter` objects as
        values.
            
    Raises:
        FileNotFoundError: If *filename* cannot be found.
        RuntimeError: If a line in *filename* cannot be parsed.
    """
    default = 'parameters.txt'
    filename = path if path else _fullpath(default)
    return _load_data(filename, 'parameter')
    

def load_errors(path=None):
    """Parse the error file and return a dictionary of 
    errors.
    
    This works like :func:`load_parameters`, but the default file is
    ``'errors.txt'``, and the returned :class:`dict` contains
    :class:`ErrorOrWarning` objects.
    """
    default = 'errors.txt'
    filename = path if path else _fullpath(default)
    return _load_data(filename, 'error')


def load_warnings(path=None):
    """Parse the warning file and return a dictionary of 
    warnings.
    
    This works like :func:`load_parameters`, but the default file is
    ``'warnings.txt'``, and the returned :class:`dict` contains
    :class:`ErrorOrWarning` objects.
    """
    default = 'warnings.txt'
    filename = path if path else _fullpath(default)
    return _load_data(filename, 'warning')


def _load_data(filename, type_):
    """Return a dictionary of parameters, errors or warnings.
    
    Args:
        filename: A text file containing parameter/error/warning data.
            The syntax used in the text file is explained in 
            ``parameters.txt``.
        type_: ``'parameters'``, ``'errors'`` or ``'warnings'``.
            
    Returns:
        A dictionary with numbers as keys and :class:`Parameter` 
        or :class:`ErrorOrWarning` objects as values.
            
    Raises:
        FileNotFoundError: If *filename* cannot be found.
        RuntimeError: If a line in *filename* cannot be parsed.
    """
    
    with open(filename, 'r') as file:        
        contents = file.read()
        
    lines = contents.split('\n')
    
    object_list = []
    for i, line in enumerate(lines):
        try:
            parsed = _parse(line, type_)
        except ValueError as e:
            raise RuntimeError(
                f'Line {i+1} of {filename} could not be parsed: ' + str(e))
        
        # *parsed* is None for empty lines; skip those.
        if parsed:
            object_list.append(parsed)
    
    return {p.number: p for p in object_list}


def _parse(line, type_):
    """Parse data from *line* and form a Parameter or ErrorOrWarning object.
    
    Args:
        line: A string with no line breaks.
        type_: 'parameter', 'error' or 'warning'.
    
    Raises:
        ValueError: If *line* cannot be parsed.
    """
    # This method is private since it's only used internally by this module,
    # but it cannot be removed/renamed without changing the tests, since a
    # lot of them use this.
    
    line = _remove_comments(line)
    
    if _is_empty(line):
        return None
    
    number_of_fields = {'parameter': 9, 'error': 4, 'warning': 4}
    fields = _separate_fields(line)
    
    if len(fields) != number_of_fields[type_]:
        raise ValueError(
            f'{len(fields)} fields instead of {number_of_fields[type_]}')
    
    fields = [_add_linebreaks(_remove_quotes(field)) for field in fields]
    
    try:
        return _form_object(fields, type_)
    except ValueError as e:
        raise ValueError('invalid values: ' + str(e)) from e


def _separate_fields(line):
    """Separate the data fields contained in *line*.
    
    Args:
        line: A string with no line breaks.
        
    Returns:
        A list containing the data fields.
        E.g. _separate_fields('1 2 "3 4"') -> ['1', '2', "3 4"].
    """
    
    start = '(?: |^)' 
    # A regex that begins a data field.
    # A space or start of string.
    
    # ?: makes groups non-capturing. Example: 
    # re.findall('(?:abc)(?:def)', 'abcdef') -> ['abcdef'], but
    # re.findall('(abc)(def)', 'abcdef') -> [('abc', 'def')]
    
    end = '(?=(?:$| ))' 
    # A regex that ends a data field.
    # A space or end of string
    
    # ?= makes a regexp non-consuming; i.e. a space can here match both 
    # the end of one word and the start of another, since the end match 
    # doesn't consume it.
    
    word = '(?:[^ "]+)'
    # A regex representing a string without any spaces.
    # Any character except space or quote, 1 or more times.
    
    many_words = '(?:"[^"]*")' 
    # A regex representing a string consisting of multiple words 
    # separated by spaces; all enclosed in double quotes.
    # Any character except quote, 1 or more times, between quotes.
    
    one_or_many_words = f'({word}|{many_words})'
    field = start + one_or_many_words + end
    return re.findall(field, line)


def _form_object(fields, type_):
    """Return a Parameter or ErrorOrWarning object.
    
    Args:
        fields: A list containing the attribute values of the 
            object as strings.
        type_: 'parameter', 'error' or 'warning'.
        
    Raises:
        ValueError: If type_ is not any of the above.
    """
    
    if type_ == 'parameter':
         return _form_parameter(fields)
    if type_ in ('error', 'warning'):
         return _form_error_or_warning(fields)
     
    raise ValueError(
        f"*type_* should be 'parameter', 'error' or 'warning', not {type_}")


def _form_parameter(fields):
    """Construct a Parameter object from given data fields.
        
    Raises:
        ValueError: If any of the data fields cannot be parsed into 
            values of the correct type.
    """    
    
    number, indices = _parse_number(  fields[0])
    name            =                 fields[1]
    unit            =                 fields[5]
    writable        = _parse_writable(fields[6])
    datatype, bits  = _parse_format(  fields[7])
    description     =                 fields[8]
        
    min_value       = _parse_minmax(  fields[2], datatype, bits)
    max_value       = _parse_minmax(  fields[3], datatype, bits)
    default         = _parse_default( fields[4], datatype, bits)
    
    return Parameter(number, name, indices, min_value, max_value, default, 
                     unit, writable, datatype, bits, description)
    

def _form_error_or_warning(fields):
    """Construct an ErrorOrWarning object from given data fields.
        
    Raises:
        ValueError: If any of the data fields cannot be parsed into 
            values of the correct type.
    """  
    number, _      = _parse_number(fields[0])
    name           =               fields[1]
    possible_cause =               fields[2]
    remedy         =               fields[3]
    
    return ErrorOrWarning(number, name, possible_cause, remedy)
        

def _parse_number(string):
    """Parse the data field containg the parameter/error/warning 
    number and possible indices.
    
    Returns:
        A tuple of the number (an int) and the indices (a range object;
        range(0) for unindexed parameters/errors/warnings).
    """
    any_number = '([0-9]+)'
    no_capture = '?:'
    one_or_zero_times = '?'
    
    index_numbers = f'\\[{any_number}:{any_number}\\]'
    # [<number>:<number>]
    maybe_index_numbers = f'({no_capture}{index_numbers}){one_or_zero_times}'
    # [<number>:<number>] or nothing
    
    numbers = f'^{any_number}{maybe_index_numbers}$'
    # <number>[<number>:<number>] or <number>

    try:
        parts = re.findall(numbers, string)[0]
        # If re.findall is successful, it will return 
        # [('<num>', '<num>', '<num>')] or [('<num>', '', '')].
        # If it's unsuccessful, it will return [].
        
    except IndexError:
        raise ValueError(f'invalid number or indices: {string}')
        
    number = int(parts[0])
    
    if parts[1] == parts[2] == '':
        indices = range(0)
    else:
        try:
            indices = range(int(parts[1]), int(parts[2]) + 1)
            # parts[2] is the last index, so 1 has to be added to it 
            # for it to be included in the range.
        except ValueError:
            raise ValueError(f'invalid indices: {string}')

    return number, indices    


def _parse_minmax(string, datatype, bits):
    """Parse a data field containg the minimum or maximum parameter
    value.
    
    Args:
        string: The field to be parsed.
        datatype: The datatype (a Data subclass) of the parameter.
        bits: The size of the parameter in bits.
    
    Returns:
        -An object of type *datatype*, if the value is numeric.
        -A string in the format P<number>, if *string* matches that
            format.    
    """
    # Try to parse the string as an int or a float first.
    try:
        return _parse_value(string, datatype, bits)
    except ValueError:
        pass
    
    # If that doesn't work, try to interpret the string as a reference.     
    any_number = '([0-9]+)'
    P_number = f'^P{any_number}$'
    
    if re.match(P_number, string):
        return string
    else:
        raise ValueError(f'invalid min/max value: {string}')    
        
        
def _parse_default(string, datatype, bits):
    """Parse the data field containg the default parameter value.
    
    Args:
        string: The field to be parsed.
        datatype: The datatype (a Data subclass) of the parameter.
        bits: The size of the parameter in bits.
    
    Returns:
        An object of type *datatype*. If different parameter indices have
        different default values, a list of such objects is returned instead.
    """
    # Try to parse the string as an int or a float first.
    try:
        return _parse_value(string, datatype, bits)
    except ValueError:
        pass
    
    # If that doesn't work, try to interpret the string as a list of
    # ints/floats.     
    try:
        value_list = ast.literal_eval(string) # Unlike eval(), this is safe.
        # By converting i to a string first we can use _parse_value to
        # e.g. parse '0' into a Float, even though Float(0) raises an error. 
        return [_parse_value(str(i), datatype, bits) for i in value_list]
    except (SyntaxError, ValueError):
        raise ValueError(f'invalid default value: {string}')
    
    
def _parse_value(string, datatype, bits):
    """Parse *string* to an int or a float and the convert that into a *bits*
    bit instance of *datatype*.
    
    Raises:
        ValueError: If *string* isn't a valid int or a float.'
    """
    # This is needed, because Uint, Sint, and Float don't accept str arguments. 
    builtin_type = float if datatype == Float else int
    value = builtin_type(string)
    return datatype(value, bits)


def _parse_format(string):
    """Parse the data field containg the parameter number format.
    
    Returns: A tuple containing the type (a subclass of Data) 
    and bits (16 or 32) of the parameter.
    """
    numbers = '([0-9]+)'
    letters = '([a-z]+)'
    letters_and_numbers = f'^{letters}{numbers}$'
    
    try:
        parts = re.findall(letters_and_numbers, string)[0]
        # If re.findall is successful, it will return 
        # [('<letters>', '<number>')].
        # If it's unsuccessful, it will return [].
        
    except IndexError:
        raise ValueError(f'invalid format: {string}')
    
    type_part = parts[0]
    bits_part = parts[1]
    
    if type_part == 'u':
        type_ = Uint
    elif type_part == 's':
        type_ = Sint
    elif type_part == 'real':
        type_ = Float
    else:
        raise ValueError(f'invalid type: {string}')
    
    try:
        bits = int(bits_part)
        
    except ValueError:
        raise ValueError(f'invalid bits: {string}')
        
    return type_, bits


def _parse_writable(string):
    """Parse the data field indicating whether the parameter can be 
    written to an return a boolean (True for writable, False otherwise).
    """    
    if string == 'r/w':
        return True
    elif string in ('r', ''):
        return False
    else:
        raise ValueError(f'invalid r/w string: {string}')


def _remove_comments(line):
    """Removes everything after a comment symbol from *line*."""
    return line.split(_COMMENT_CHAR)[0]
            
          
def _is_empty(line):
    """Returns True if *line* consists only of whitespace."""
    empty_regex = '^\\s*$' # \s = whitespace character
    return bool(re.match(empty_regex, line))
              
        
def _remove_quotes(string):
    """If *string* is enclosed in double quotes, this function removes 
    them.
    
    E.g. both 'Lorem ipsum' and '"Lorem ipsum"' return 'Lorem ipsum'.
    """
    
    if string[0] == string[-1] == '"':
        return string[1:-1]
    else:
        return string


def _add_linebreaks(string):
    """Replaces occurrences of r'\n' in *string* with '\n'.
    
    Line breaks can be manually inserted inside data fields in a 
    text file by typing a backslash and a 'n'. These are two 
    separate characters, and need to be replaced by a single 
    line break character '\n'.
    """    
    
    # The Python interpreter and the re module seem to apply special
    # characters separately, so we need to escape the backslash 
    # character twice. 
    # r'\\' equals two backslashes (r denotes a raw string, where no 
    # special characters are applied) and the re module interprets 
    # this as a single backslash. 
    # '\\\\' also works, since it is equal to r'\\'.
    
    # Here r'\\n' is interpreted by the re module as a single 
    # backslash and the letter n.
    return re.sub(r'\\n', '\n', string)
           
 
main()



"""Unit tests for the parameter_component module."""

import unittest

from turboctl.telegram.codes import (AccessError, MinMaxError,
                                     ParameterIndexError, WrongNumError)
from turboctl.telegram.datatypes import Uint, Sint, Float
from turboctl.telegram.telegram import TelegramBuilder, TelegramReader
from turboctl.virtualpump.parameter_component import (ExtendedParameters,
                                                      ParameterComponent)
from test_turboctl.telegram.test_parser import dummy_parameter


class Base(unittest.TestCase):
    
    ### Set-up ###

    def setUp(self):
        self.parameters = self.set_up_parameters()
        self.extended_parameters = ExtendedParameters(self.parameters)
        self.pc = ParameterComponent(self.extended_parameters)
        
    def set_up_parameters(self):
        # Testing is easier with a custom parameter set instead of actual pump
        # parameters.
        
        def dp(number, datatype, bits, indexed, min_value, max_value, default):
            return dummy_parameter(
                number=number,
                indices=range(1,4) if indexed else range(0), 
                datatype=datatype,
                bits=bits,
                min_value=datatype(min_value, bits),
                max_value=datatype(max_value, bits),
                default=datatype(default, bits)
            )
        
        self.u16  = dp(1, Uint, 16, False, 10, 10**4, 123)
        self.u32  = dp(2, Uint, 32, False, 10, 10**8, 123456)
        self.u16F = dp(3, Uint, 16, True,  10, 10**4, 123)
        self.u32F = dp(4, Uint, 32, True,  10, 10**8, 123456)
        
        self.s16  = dp(5, Sint, 16, False, -10**4, 10**4, -123)
        self.s32  = dp(6, Sint, 32, False, -10**8, 10**8, -123456)
        self.s16F = dp(7, Sint, 16, True,  -10**4, 10**4, -123)
        self.s32F = dp(8, Sint, 32, True,  -10**8, 10**8, -123456)
        
        self.real  = dp( 9, Float, 32, False, -1e8, 1e8, 123.456)
        self.realF = dp(10, Float, 32, True,  -1e8, 1e8, 123.456)
        
        parameters = [self.u16, self.u32, self.u16F, self.u32F, self.s16, 
                      self.s32, self.s16F, self.s32F, self.real, self.realF]
        return {p.number: p for p in parameters}    
        
    ### Generic testing functionality ###
                
    def access_parameter(self, number, mode, value, index):
        if index == None:
            index = 2 if self.parameters[number].indices else 0
 
        builder = (TelegramBuilder(self.parameters)
                   .set_parameter_number(number)
                   .set_parameter_mode(mode)
                   .set_parameter_index(index)
                   .set_parameter_value(value))
            
        query = TelegramReader(builder.build('query'), 'query')
        # The reply initially contains the same data as the query, so we can
        # use the builder used to build the query as a basis for the reply.
        self.pc.handle_parameter(query, builder)
        reply = TelegramReader(builder.build('reply'), 'reply')
        
        return reply
    
    def check_response(self, reply, number, mode, value, index):
        if index == None:
            index = 2 if self.parameters[number].indices else 0
        
        self.assertEqual(reply.parameter_mode, mode)
        self.assertEqual(reply.parameter_number, number)
        self.assertEqual(reply.parameter_index, index)
        # assertAlmostEqual is used because the use of floats may cause a
        # slight difference between the values.
        self.assertAlmostEqual(reply.parameter_value, value, 5)
        
    def check_error(self, reply, number, error_class):
        self.assertEqual(reply.parameter_mode, 'error')
        self.assertEqual(reply.parameter_number, number)
        self.assertEqual(reply.parameter_error, error_class.member)
        
    ### Generic tests for different purposes ###
        
    def read_test(self, number, value, index=None):
        value_in = 0.0 if self.parameters[number].datatype == Float else 0
        reply = self.access_parameter(number, 'read', value_in, index)
        self.check_response(reply, number, 'response', value, index)
    
    def write_test(self, number, value, index=None):
        reply = self.access_parameter(number, 'write', value, index)
        self.check_response(reply, number, 'response', value, index)
        
    def no_write_access_test(self, number, value, index=None):
        reply = self.access_parameter(number, 'write', value, index)
        self.check_response(reply, number, 'no write', value, index)
        
    def write_error_test(self, number, value, error_class, index=None):
        reply = self.access_parameter(number, 'write', value, index)
        self.check_error(reply, number, error_class)
        
    def read_error_test(self, number, value, error_class, index=None):
        reply = self.access_parameter(number, 'read', value, index)
        self.check_error(reply, number, error_class)
        
        
class TestDefaults(Base):
    """Make sure ExtendedParameter objects set their default values correctly.
    """

    def set_up_parameters(self):
        
        def dp(number, indexed, default):
            indices = range(1,4) if indexed else range(0)
            return dummy_parameter(number=number, indices=indices, 
                                   default=default)
                   
        self.unindexed = dp(1, False, Uint(1, 16))
        self.single = dp(2, True, Uint(2, 16))
        self.multiple = dp(3, True, [Uint(3, 16), Uint(4, 16), Uint(5, 16)])
        
        parameters = [self.unindexed, self.single, self.multiple]        
        return {p.number: p for p in parameters}        

    def test_unindexed(self):
        """Test an unindexed parameter."""
        self.assertEqual(self.pc.parameters[1].value, [Uint(1, 16)])
        self.read_test(self.unindexed.number, 1)
        
    def test_indexed_one(self):
        """Test an indexed parameter with one default value."""
        values = 3 * [Uint(2, 16)]
        self.assertEqual(self.pc.parameters[2].value, values)
        self.read_test(self.single.number, 2)
        
    def test_indexed_multiple(self):
        """Test an indexed parameter with multiple default values."""
        values = [Uint(x, 16) for x in (3, 4, 5)]
        self.assertEqual(self.pc.parameters[3].value, values)
        self.read_test(self.multiple.number, value=3, index=1)
        self.read_test(self.multiple.number, value=4, index=2)
        self.read_test(self.multiple.number, value=5, index=3)
        

class TestReadWrite(Base):
    """Test the basic reading and writing functionality of ParameterComponent.
    """
    
    def test_no_access(self):
        # If the partameter isn't accessed, the reply will just be a copy of
        # the query.
        reply = self.access_parameter(0, 'none', 12, 34)
        self.check_response(reply, 0, 'none', 12, 34)
                    
    def test_read_defaults(self):
        for parameter in self.parameters.values():
            with self.subTest(i=parameter.number):
                self.read_test(parameter.number, parameter.default.value)

    def test_write_and_read_single_value(self):
        
        def half(value):
            return type(value)(value / 2)
        
        # Write and read a value for all parameters.
        for parameter in self.parameters.values():
            with self.subTest(i=parameter.number):
                value = half(parameter.default.value)
                self.write_test(parameter.number, value)
                self.read_test(parameter.number, value)
                
    def test_write_and_read_all_indices(self):
        
        def half(value):
            return type(value)(value / 2)
        
        # Write and read a value for all parameters.
        for parameter in self.parameters.values():
            if parameter.indices:
                with self.subTest(i=parameter.number):
                    
                    value1 = half(parameter.default.value)
                    value2 = 2 * parameter.default.value
                    value3 = 4 * parameter.default.value
                    
                    self.write_test(parameter.number, value1, index=1)
                    self.write_test(parameter.number, value2, index=2)
                    self.write_test(parameter.number, value3, index=3)
                    
                    self.read_test(parameter.number, value1, index=1)
                    self.read_test(parameter.number, value2, index=2)
                    self.read_test(parameter.number, value3, index=3)
                
    # TODO: This needs testing with the real pump to implement correctly.
    # def test_no_write_access(self):
    #     # A 'no write' response isn't an error, so it's tested here.
    #     for parameter in self.parameters.values():
    #         self.pc.parameters[parameter.number].writable = False
    #         with self.subTest(i=parameter.number):                    
    #             self.no_write_access_test(parameter.number,
    #                                       parameter.default.value)
                

class TestLimits(Base):
    """Make sure min_value and max_value work correctly when they are costant
    values.
    """
                
    def test_value_too_large(self):
        for parameter in self.parameters.values():
            with self.subTest(i=parameter.number):
                value = 2 * parameter.max_value.value
                self.write_error_test(parameter.number, value, MinMaxError)
                
    def test_value_too_small(self):
        for parameter in self.parameters.values():
            with self.subTest(i=parameter.number):
                min_val = parameter.min_value.value
                value = min_val - abs(min_val)
                self.write_error_test(parameter.number, value, MinMaxError)
                
    def test_maximum_value(self):
        for parameter in self.parameters.values():
            with self.subTest(i=parameter.number):
                value = parameter.max_value.value
                self.write_test(parameter.number, value)
                
    def test_minimum_value(self):
        for parameter in self.parameters.values():
            with self.subTest(i=parameter.number):
                value = parameter.min_value.value
                self.write_test(parameter.number, value)
                
                
class TestLimitsAsReferences(Base):
    """Make sure min_value and max_value also work correctly when they are
    references to other parameters.
    """
    
    def set_up_parameters(self):
        
        def dp(number, min_value, max_value, default):
            if not isinstance(min_value, str):
                min_value = Sint(min_value, 16)
            
            if not isinstance(max_value, str):
                max_value = Sint(max_value, 16)
            
            return dummy_parameter(number=number, datatype=Sint,
                                   min_value=min_value, max_value=max_value,
                                   default=Sint(default, 16))
                   
        self.lower = dp(  1, -100, 100, -1)
        self.upper = dp(100, -100, 100,  1)
        # The number is 100 in order to make sure references can be 
        # more than 1 digit long.
        self.middle = dp(5, 'P1', 'P100', 0)        
        parameters = [self.lower, self.middle, self.upper]        
        return {p.number: p for p in parameters}

    def test_min_value_as_reference(self):
        self.assertEqual(self.pc.parameters[5].min_value, Sint(-1, 16))
        self.write_test(self.lower.number, -2)
        self.assertEqual(self.pc.parameters[5].min_value, Sint(-2, 16))
        
    def test_max_value_as_reference(self):
        self.assertEqual(self.pc.parameters[5].max_value, Sint(1, 16))
        self.write_test(self.upper.number, 2)
        self.assertEqual(self.pc.parameters[5].max_value, Sint(2, 16))
        
    def test_variable_min_and_max_value(self):
        
        # The value 5 is out of range.
        self.write_error_test(self.middle.number, 5, MinMaxError)
        # Raise the maximum value.
        self.write_test(self.upper.number, 5)
        # The value is now allowed.
        self.write_test(self.middle.number, 5)
        
        # The same for the minimum value.
        self.write_error_test(self.middle.number, -5, MinMaxError)
        self.write_test(self.lower.number, -5)
        self.write_test(self.middle.number, -5)


class TestStr(Base):
    """Test ExtendedParameter.__str__."""
    
    def set_up_parameters(self):
        # Copy the parameters from TestLimitsAsReferences.set_up_parameters,
        # since that class tests references too.
        return TestLimitsAsReferences.set_up_parameters(self)
        
    def test_str(self):
        correct_string = """
ExtendedParameter(
    number=5,   
    indices=range(0, 0),
    min_value='P1' (Sint(-1, bits=16)),
    max_value='P100' (Sint(1, bits=16)),
    default=Sint(0, bits=16),
    writable=True,
    datatype=Sint,
    bits=16,
    value=[Sint(0, bits=16)]
)
"""[1:-1]
        string = str(self.pc.parameters[self.middle.number])
        self.assertEqual(string, correct_string)
        
    def test_str_with_references(self):
        correct_string = """
ExtendedParameter(
    number=1,   
    indices=range(0, 0),
    min_value=Sint(-100, bits=16),
    max_value=Sint(100, bits=16),
    default=Sint(-1, bits=16),
    writable=True,
    datatype=Sint,
    bits=16,
    value=[Sint(-1, bits=16)]
)
"""[1:-1]
        string = str(self.pc.parameters[self.lower.number])
        self.assertEqual(string, correct_string)


class TestInvalidNumber(Base):
    """Make sure that an invalid parameter number results in the correct
    error code.
    """
    
    def setUp(self):
        super().setUp()
        # TelegramBuilder usually raises an error if it is told to build a
        # telegram that accesses an invalid parameter.
        # This can be circumvented by using different parameter dictionaries in
        # TelegramBuilder and ParameterComponent.
        new_parameters = {0: dummy_parameter(number=0)} 
        self.builder = TelegramBuilder(new_parameters).set_parameter_number(0)
    
    def test_read_invalid_parameter_number(self):  
        self.builder.set_parameter_mode('read')
        
        query = TelegramReader(self.builder.build('query'), 'query')
        self.pc.handle_parameter(query, self.builder)
        reply = TelegramReader(self.builder.build('reply'), 'reply')
        
        self.assertEqual(reply.parameter_error.exception, WrongNumError)
        
    def test_write_invalid_parameter_number(self):  
        self.builder.set_parameter_mode('write')
        
        query = TelegramReader(self.builder.build('query'), 'query')
        self.pc.handle_parameter(query, self.builder)
        reply = TelegramReader(self.builder.build('reply'), 'reply')
        
        self.assertEqual(reply.parameter_error.exception, WrongNumError)

                                
class TestWrongBits(Base):
    """Make sure that accessing a 16-bit parameter with a 32-bit mode or vice
    versa results in the correct error code.
    """
    
    def setUp(self):
        super().setUp()
        for parameter in self.pc.parameters.values():
            parameter.bits = 32 if parameter.bits == 16 else 16

    def test_write_wrong_size(self):
        for parameter in self.parameters.values():
            with self.subTest(i=parameter.number):
                value = parameter.default.value
                self.write_error_test(parameter.number, value, AccessError)
                
    def test_read_wrong_size(self):
        # Read access doesn't specify size, so reading parameters 
        # with a wrong size should succeed.
        for parameter in self.parameters.values():
            with self.subTest(i=parameter.number):
                value = parameter.default.value
                self.read_test(parameter.number, value)
                
                
class TestWrongIndexed(Base):
    """Make sure that accessing an indexed parameter with an unindexed mode
    or vice versa results in the correct error code.
    """
    
    def setUp(self):
        super().setUp()
        # TODO: This might change if the implementation is changed
        for parameter in self.pc.parameters.values():
            parameter.indices = range(0) if parameter.indices else range(5)

    def test_wrong_indexed(self):
        for parameter in self.parameters.values():
            with self.subTest(i=parameter.number):
                value = parameter.default.value
                self.write_error_test(parameter.number, value, AccessError)
                
                
class TestIndexError(Base):
    """Make sure that accessing an invalid index of a parameter results in the
    correct error code.
    """
    
    def test_read_write_index_too_large(self):
        for parameter in self.parameters.values():
            if parameter.indices:
                with self.subTest(i=parameter.number):
                    value = parameter.default.value
                    index = 5
                    self.read_error_test(
                        parameter.number, value, ParameterIndexError, index)
                    self.write_error_test(
                        parameter.number, value, ParameterIndexError, index)
                    
    def test_read_write_index_too_small(self):
        for parameter in self.parameters.values():
            if parameter.indices:
                with self.subTest(i=parameter.number):
                    value = parameter.default.value
                    index = 0
                    self.read_error_test(
                        parameter.number, value, ParameterIndexError, index)
                    self.write_error_test(
                        parameter.number, value, ParameterIndexError, index)
                    

if __name__ == '__main__':
    unittest.main()


"""Unit tests for the virtualconnection module."""

import unittest
import os
from serial import Serial

from turboctl.virtualpump.virtualconnection import VirtualConnection


def process(input_):
    return b'reply to ' + input_

    
class Base(unittest.TestCase):
    
    def setUp(self):
        self.vc = VirtualConnection(process)
        
    def tearDown(self):
        self.vc.close()


class TestMessaging(Base):
    
    def setUp(self):
        super().setUp()
        self.vc.buffer_size = 6
        self.ser = Serial(port=self.vc.port, timeout=0.1)
        
    def tearDown(self):
        super().tearDown()
        self.ser.close()
        
    def test_message_smaller_than_vc_buffer(self):
        
        msg = b'hello'
        self.ser.write(msg)
        
        reply1 = self.ser.read(100)
        self.assertEqual(reply1, b'reply to hello')
        
        reply2 = self.ser.read(100)
        self.assertFalse(reply2)
            
    def test_message_larger_than_vc_buffer(self):
        
        msg = b'hello there'
        self.ser.write(msg)
        
        reply = self.ser.read(100)
        self.assertEqual(reply, b'reply to hello reply to there')
        
    def test_message_larger_than_vc_buffer_and_read_in_pieces(self):
        
        msg = b'hello there'
        self.ser.write(msg)
        
        reply = self.ser.read(15)
        self.assertEqual(reply, b'reply to hello ')
        
        reply = self.ser.read(15)
        self.assertEqual(reply, b'reply to there')
        
        reply = self.ser.read(15)
        self.assertEqual(reply, b'')
        
    def test_multiple_messages(self):
        
        self.ser.write(b'hello')
        self.assertEqual(self.ser.read(100), b'reply to hello')
        
        self.ser.write(b'there')
        self.assertEqual(self.ser.read(100), b'reply to there')
                
        self.assertEqual(self.ser.read(100), b'')

        
class TestStartingAndStopping(Base):
        
    def test_init(self):
        self.assertTrue(self.vc.is_running)
        self.assertTrue(os.fstat(self.vc.user_end))
        self.assertTrue(os.fstat(self.vc.virtual_end))
        
    def test_close(self):
        self.vc.close()
        self.assertFalse(self.vc.is_running())
        
        # Asking the status of user_end and virtual_end should 
        # produce an error, since they are no longer valid file 
        # descriptors after they are closed.  
        with self.assertRaises(OSError):
            os.fstat(self.vc.user_end)
        with self.assertRaises(OSError):
            os.fstat(self.vc.virtual_end)
            
    def test_close_after_close(self):
        """Make sure closing a VC twice doesn't raise errors."""
        self.vc.close()
        self.vc.close()
                    
    def test_close_all(self):
        vc1 = self.vc
        vc2 = VirtualConnection(process)
        vc3 = VirtualConnection(process)
        self.assertTrue(vc1.is_running())
        self.assertTrue(vc2.is_running())
        self.assertTrue(vc3.is_running())
        
        VirtualConnection.close_all()
        self.assertFalse(vc1.is_running())
        self.assertFalse(vc2.is_running())
        self.assertFalse(vc3.is_running())
        
    def test_init_after_close_all(self):
        VirtualConnection.close_all()
        with VirtualConnection() as vc:
            self.assertTrue(vc.is_running)
            self.assertTrue(os.fstat(vc.user_end))
            self.assertTrue(os.fstat(vc.virtual_end))
        
    def test_running_instances(self):
        VirtualConnection.close_all()
        
        self.assertEqual(len(VirtualConnection.running_instances), 0)
        vc1 = VirtualConnection(process)
        self.assertEqual(len(VirtualConnection.running_instances), 1)
        VirtualConnection(process)
        self.assertEqual(len(VirtualConnection.running_instances), 2)
        VirtualConnection(process)
        self.assertEqual(len(VirtualConnection.running_instances), 3)

        vc1.close()
        self.assertEqual(len(VirtualConnection.running_instances), 2)
        VirtualConnection.close_all()
        self.assertEqual(len(VirtualConnection.running_instances), 0)
        
        
if __name__ == '__main__':
    unittest.main()